from typing import Iterable, Optional, Tuple, Union

from PIL import (
    ImageDraw as PILImageDraw,
    ImageColor as PILImageColor,
)
import numpy as np
import cv2 as cv

from .image import Image
from .point import Point, PointList
from .line import Line
from .box import Box
from .polygon import Polygon, TextPolygon
from .mask import Mask
from .score_map import ScoreMap

# https://mokole.com/palette.html
_DISTINCT_HEX_RGB_COLORS = [
    # darkgreen
    '#006400',
    # darkblue
    '#00008b',
    # maroon3
    '#b03060',
    # red
    '#ff0000',
    # yellow
    '#ffff00',
    # burlywood
    '#deb887',
    # lime
    '#00ff00',
    # aqua
    '#00ffff',
    # fuchsia
    '#ff00ff',
    # cornflower
    '#6495ed',
]


def get_rgb_tuple(hex_color: str) -> Tuple[int, int, int]:
    # https://pillow.readthedocs.io/en/stable/reference/ImageColor.html#color-names
    return PILImageColor.getrgb(hex_color)  # type: ignore


def get_complementary_rgb_or_rgba_color(
    color: Union[Tuple[int, int, int], Tuple[int, int, int, int]],
):
    complementary_color = (255 - color[0], 255 - color[1], 255 - color[2])
    if len(color) == 4:
        complementary_color = (*complementary_color, color[-1])
    return complementary_color


def get_color(
    hex_color: str,
    distinct_idx: Optional[int] = None,
    alpha: Optional[int] = None,
):
    if hex_color == 'distinct':
        assert distinct_idx is not None
        hex_color = _DISTINCT_HEX_RGB_COLORS[distinct_idx % len(_DISTINCT_HEX_RGB_COLORS)]

    rgb = get_rgb_tuple(hex_color)
    if alpha:
        # RGBA.
        return (*rgb, alpha)
    else:
        # RGB.
        return rgb


def visualize_points(
    image: Image,
    points: Union[PointList, Iterable[Point]],
    style: str = 'circle-1',
    fill: str = 'distinct',
):
    pil_image = image.to_rgb_image().to_pil_image()
    draw = PILImageDraw.Draw(pil_image)

    if style == 'pixel':
        for idx, point in enumerate(points):
            draw.point([(point.x, point.y)], fill=get_color(fill, idx))

    elif style.startswith('circle-'):
        radius = int(style[len('circle-'):])
        for idx, point in enumerate(points):
            up = max(point.y - radius, 0)
            down = min(point.y + radius, image.height - 1)
            left = max(point.x - radius, 0)
            right = min(point.x + radius, image.width - 1)
            draw.ellipse((left, up, right, down), fill=get_color(fill, idx))

    else:
        raise NotImplementedError()

    return Image.from_pil_image(pil_image)


def visualize_lines(
    image: Image,
    lines: Iterable[Line],
    fill: str = 'distinct',
    show_index: bool = False,
):
    pil_image = image.to_rgb_image().to_pil_image()
    draw = PILImageDraw.Draw(pil_image)

    for idx, line in enumerate(lines):
        color = get_color(fill, idx)
        draw.line(line.to_xy_pairs(), fill=color)
        if show_index:
            draw.text(line.get_center_point().to_xy_pair(), str(idx), fill=color)

    return Image.from_pil_image(pil_image)


def visualize_polygons(
    image: Image,
    polygons: Iterable[Polygon],
    fill: Optional[str] = 'distinct',
    outline: Optional[str] = 'distinct',
    alpha: int = 128,
    show_index: bool = False,
    index_color: str = 'distinct',
):
    pil_image = image.to_rgba_image().to_pil_image()
    pil_polygon_image = Image.from_shapable(image, num_channels=4, fill_value=0).to_pil_image()
    draw = PILImageDraw.Draw(pil_polygon_image)

    for idx, polygon in enumerate(polygons):
        fill_color = None
        outline_color = None

        if fill:
            fill_color = get_color(fill, idx, alpha)
            outline_color = fill_color

        if outline:
            outline_color = get_color(outline, idx, alpha if fill else 255)

        assert fill_color or outline_color
        draw.polygon(tuple(polygon.to_flatten_xy_pairs()), fill=fill_color, outline=outline_color)

        if show_index:
            text_fill = fill_color or outline_color
            assert text_fill and index_color

            if index_color == 'distinct':
                text_fill = get_complementary_rgb_or_rgba_color(text_fill)
            else:
                text_fill = get_color(index_color, idx, alpha if fill else 255)

            y_mean = round(np.mean([point.y for point in polygon.points]))
            x_mean = round(np.mean([point.x for point in polygon.points]))

            draw.text((x_mean, y_mean), str(idx), fill=text_fill)

    pil_image.alpha_composite(pil_polygon_image)
    return Image.from_pil_image(pil_image)


def visualize_text_polygons(
    image: Image,
    text_polygons: Iterable[TextPolygon],
    fill: Optional[str] = 'distinct',
    outline: Optional[str] = 'distinct',
    alpha: int = 128,
    show_index: bool = False,
    index_color: str = 'distinct',
):
    return visualize_polygons(
        image=image,
        polygons=[text_polygon.polygon for text_polygon in text_polygons],
        fill=fill,
        outline=outline,
        alpha=alpha,
        show_index=show_index,
        index_color=index_color,
    )


def visualize_boxes(
    image: Image,
    boxes: Iterable[Box],
    fill: Optional[str] = 'distinct',
    outline: Optional[str] = 'distinct',
    alpha: int = 128,
    show_index: bool = False,
    index_color: str = 'distinct',
):
    return visualize_polygons(
        image=image,
        polygons=(box.to_polygon() for box in boxes),
        fill=fill,
        outline=outline,
        alpha=alpha,
        show_index=show_index,
        index_color=index_color,
    )


def overlay_boxes_of_image(
    src_image: Image,
    boxes: Iterable[Box],
    dst_image: Image,
):
    dst_image = dst_image.copy()
    for box in boxes:
        boxed_image = box.extract_image(src_image)
        box.fill_image(dst_image, boxed_image)
    return dst_image


def visualize_score_map(
    score_map: ScoreMap,
    cv_colormap: int = cv.COLORMAP_JET,
    enable_boundary_equalization: bool = False,
    enable_center_shift: bool = False,
):
    mat = score_map.mat.copy()

    if enable_boundary_equalization:
        # Equalize to [0, 255]
        val_min = np.min(mat)
        mat -= val_min
        val_max = np.max(mat)
        mat *= 255.0
        mat /= val_max

    elif enable_center_shift:
        mat *= 127.5

    mat = np.clip(mat, 0, 255).astype(np.uint8)

    # Apply color map.
    color_mat = cv.applyColorMap(mat, cv_colormap)
    color_mat = cv.cvtColor(color_mat, cv.COLOR_BGR2RGB)

    return Image(mat=color_mat)


def visualize_mask(mask: Mask):
    new_mat = np.zeros_like(mask.mat)
    np.putmask(new_mat, mask.np_mask, 255)
    return Image(mat=new_mat)
