from typing import (
    cast,
    get_args,
    Generic,
    Type,
    TypeVar,
    Dict,
    Sequence,
    Any,
    Tuple,
    List,
    Optional,
    Union,
)
from enum import Enum, unique

import attrs
from numpy.random import RandomState
import iolite as io

from vkit.utility import rnd_choice, is_path_type, dyn_structure, PathType

T_CONFIG = TypeVar('T_CONFIG')
T_RES = TypeVar('T_RES')
T_RUN = TypeVar('T_RUN')
T_OUT = TypeVar('T_OUT')


@attrs.define
class NoneTypeEngineConfig:
    pass


@attrs.define
class NoneTypeEngineResource:
    pass


class Engine(Generic[T_CONFIG, T_RES, T_RUN, T_OUT]):

    @classmethod
    def get_type_name(cls) -> str:
        raise NotImplementedError()

    def __init__(
        self,
        config: T_CONFIG,
        resource: Optional[T_RES] = None,
    ):
        self.config = config
        self.resource = resource

    def run(self, config: T_RUN, rnd: RandomState) -> T_OUT:
        raise NotImplementedError()


class EngineRunner(Generic[T_CONFIG, T_RES, T_RUN, T_OUT]):

    def __init__(self, engine: Engine[T_CONFIG, T_RES, T_RUN, T_OUT]) -> None:
        self.engine = engine

    def get_run_config_cls(self) -> Type[T_RUN]:
        return get_args(self.engine.__orig_bases__[0])[2]  # type: ignore

    def run(self, config: Union[Dict[str, Any], T_RUN], rnd: RandomState) -> T_OUT:
        config = dyn_structure(config, self.get_run_config_cls())
        return self.engine.run(config, rnd)


class EngineFactory(Generic[T_CONFIG, T_RES, T_RUN, T_OUT]):

    def __init__(self, engine_cls: Type[Engine[T_CONFIG, T_RES, T_RUN, T_OUT]]) -> None:
        self.engine_cls = engine_cls

    def get_type_name(self):
        return self.engine_cls.get_type_name()

    def get_config_cls(self) -> Type[T_CONFIG]:
        return get_args(self.engine_cls.__orig_bases__[0])[0]  # type: ignore

    def get_resource_cls(self) -> Type[T_RES]:
        return get_args(self.engine_cls.__orig_bases__[0])[1]  # type: ignore

    def create(
        self,
        config: Optional[Union[Dict[str, Any], PathType, T_CONFIG]] = None,
        resource: Optional[Union[Dict[str, Any], T_RES]] = None,
    ):
        config = dyn_structure(
            config,
            self.get_config_cls(),
            support_path_type=True,
            support_none_type=True,
        )

        resource_cls = self.get_resource_cls()
        if resource_cls is NoneTypeEngineResource:
            assert resource is None
        else:
            assert resource
        if resource is not None:
            resource = dyn_structure(resource, resource_cls)

        return EngineRunner(self.engine_cls(config, resource))


class EngineRunnerAggregator(Generic[T_RES, T_RUN, T_OUT]):

    def __init__(
        self,
        pairs: Sequence[Tuple[EngineRunner[Any, T_RES, T_RUN, T_OUT], float]],
    ):
        self.engine_runners: List[EngineRunner[Any, T_RES, T_RUN, T_OUT]] = []
        weights = []
        for engine_runner, weight in pairs:
            self.engine_runners.append(engine_runner)
            weights.append(weight)
        total = sum(weights)
        self.probs = [weight / total for weight in weights]

    def run(self, config: Union[Dict[str, Any], T_RUN], rnd: RandomState) -> T_OUT:
        engine_runner = rnd_choice(rnd, self.engine_runners, probs=self.probs)
        return engine_runner.run(config, rnd)


@unique
class EngineRunnerAggregatorFactoryConfigKey(Enum):
    TYPE = 'type'
    WEIGHT = 'weight'
    CONFIG = 'config'


class EngineRunnerAggregatorFactory(Generic[T_RES, T_RUN, T_OUT]):

    def __init__(self, engine_factories: Sequence[EngineFactory[Any, T_RES, T_RUN, T_OUT]]):
        self.type_name_to_engine_factory = {
            engine_factory.get_type_name(): engine_factory for engine_factory in engine_factories
        }

    def get_resource_cls(self) -> Type[T_RES]:
        engine_cls = next(iter(self.type_name_to_engine_factory.values()))
        return engine_cls.get_resource_cls()

    def create(
        self,
        configs: Union[Sequence[Dict[str, Any]], PathType],
        resource: Optional[Union[Dict[str, Any], T_RES]] = None,
    ):
        resource_cls = self.get_resource_cls()
        if resource_cls is NoneTypeEngineResource:
            assert resource is None
        else:
            assert resource
        if resource is not None:
            resource = dyn_structure(resource, resource_cls)

        if is_path_type(configs):  # type: ignore
            configs = io.read_json(configs, expandvars=True)  # type: ignore
        configs = cast(Sequence[Dict[str, Any]], configs)

        pairs: List[Tuple[EngineRunner[Any, T_RES, T_RUN, T_OUT], float]] = []
        for config in configs:
            type_name = config[EngineRunnerAggregatorFactoryConfigKey.TYPE.value]
            if type_name not in self.type_name_to_engine_factory:
                raise KeyError(f'type_name={type_name} not found')

            engine_factory = self.type_name_to_engine_factory[type_name]
            engine_runner = engine_factory.create(
                config.get(EngineRunnerAggregatorFactoryConfigKey.CONFIG.value, {}),
                resource,
            )
            if len(configs) == 1:
                weight = 1
            else:
                weight = config[EngineRunnerAggregatorFactoryConfigKey.WEIGHT.value]

            pairs.append((engine_runner, weight))
        return EngineRunnerAggregator(pairs)
