from typing import get_args, Generic, TypeVar, Type, Dict, Any, Sequence, Optional, Union

import attrs
from numpy.random import RandomState

from vkit.utility import dyn_structure, PathType

T_VALUE = TypeVar('T_VALUE')


@attrs.define
class PipelineState:
    key_to_value: Dict[str, Any] = attrs.field(factory=dict)

    def get_value(self, key: str, value_cls: Type[T_VALUE]) -> T_VALUE:
        if key not in self.key_to_value:
            raise KeyError(f'key={key} not found.')
        value = self.key_to_value[key]
        if not isinstance(value, value_cls):
            raise TypeError(f'key={key}, value type={type(value)} is not instance of {value_cls}')
        return value

    def set_value(self, key: str, value: Any, override: bool = False):
        if key in self.key_to_value and not override:
            raise KeyError(f'key={key} exists but override is not set.')
        self.key_to_value[key] = value


@attrs.define
class NoneTypePipelineStepConfig:
    pass


@attrs.define
class NoneTypePipelineStepResource:
    pass


T_CONFIG = TypeVar('T_CONFIG')
T_RES = TypeVar('T_RES')


class PipelineStep(Generic[T_CONFIG, T_RES]):

    def __init__(
        self,
        config: T_CONFIG,
        resource: Optional[T_RES] = None,
    ):
        self.config = config
        self.resource = resource

    def run(self, state: PipelineState, rnd: RandomState):
        raise NotImplementedError()


class PipelineStepFactory(Generic[T_CONFIG, T_RES]):

    def __init__(self, pipeline_step_cls: Type[PipelineStep[T_CONFIG, T_RES]]):
        self.pipeline_step_cls = pipeline_step_cls

    def get_config_cls(self) -> Type[T_CONFIG]:
        return get_args(self.pipeline_step_cls.__orig_bases__[0])[0]  # type: ignore

    def get_resource_cls(self) -> Type[T_RES]:
        return get_args(self.pipeline_step_cls.__orig_bases__[0])[1]  # type: ignore

    def create(
        self,
        config: Optional[Union[Dict[str, Any], PathType, T_CONFIG]] = None,
        resource: Optional[Union[Dict[str, Any], T_RES]] = None,
    ):
        config = dyn_structure(
            config,
            self.get_config_cls(),
            support_path_type=True,
            support_none_type=True,
        )

        resource_cls = self.get_resource_cls()
        if resource_cls is NoneTypePipelineStepResource:
            assert resource is None
        else:
            assert resource
        if resource is not None:
            resource = dyn_structure(resource, resource_cls)

        return self.pipeline_step_cls(config, resource)


@attrs.define
class NoneTypePipelinePostProcessorConfig:
    pass


T_OUTPUT = TypeVar('T_OUTPUT')


class PipelinePostProcessor(Generic[T_CONFIG, T_OUTPUT]):

    def __init__(self, config: T_CONFIG):
        self.config = config

    def generate_output(
        self,
        state: PipelineState,
        rnd: RandomState,
    ) -> T_OUTPUT:
        raise NotImplementedError()


class BypassPipelinePostProcessor(
    PipelinePostProcessor[NoneTypePipelinePostProcessorConfig, PipelineState]
):

    def generate_output(self, state: PipelineState, rnd: RandomState) -> PipelineState:
        return state


class PipelinePostProcessorFactory(Generic[T_CONFIG, T_OUTPUT]):

    def __init__(
        self, pipeline_post_processor_cls: Type[PipelinePostProcessor[T_CONFIG, T_OUTPUT]]
    ):
        self.pipeline_post_processor_cls = pipeline_post_processor_cls

    def get_config_cls(self) -> Type[T_CONFIG]:
        return get_args(self.pipeline_post_processor_cls.__orig_bases__[0])[0]  # type: ignore

    def create(
        self,
        config: Optional[Union[Dict[str, Any], PathType, T_CONFIG]] = None,
    ):
        config = dyn_structure(
            config,
            self.get_config_cls(),
            support_path_type=True,
            support_none_type=True,
        )
        return self.pipeline_post_processor_cls(config)


bypass_post_processor_factory = PipelinePostProcessorFactory(BypassPipelinePostProcessor)


class Pipeline(Generic[T_OUTPUT]):

    def __init__(
        self,
        steps: Sequence[PipelineStep],
        post_processor: PipelinePostProcessor[Any, T_OUTPUT],
    ):
        self.steps = steps
        self.post_processor = post_processor

    def run(
        self,
        rnd: RandomState,
        state: Optional[PipelineState] = None,
    ) -> T_OUTPUT:
        if state is None:
            state = PipelineState()
        for step in self.steps:
            step.run(state, rnd)
        return self.post_processor.generate_output(state, rnd)
