import attrs
from numpy.random import RandomState

from vkit.element import Image, Shapable
from vkit.pipeline.interface import (
    PipelineStep,
    NoneTypePipelineStepResource,
    PipelineStepFactory,
    PipelineState,
)
from vkit.pipeline.step.page_layout import PageLayout
from vkit.pipeline.step.page_image import PageImageCollection
from vkit.pipeline.step.page_text_line import PageTextLineCollection


@attrs.define
class PageAssemblerStepConfig:
    state_key_page_layout: str = 'page_layout'
    state_key_background_image: str = 'background_image'
    state_key_page_image_collection: str = 'page_image_collection'
    state_key_page_text_line_collection: str = 'page_text_line_collection'
    state_key_page: str = 'page'


@attrs.define
class Page(Shapable):
    image: Image
    page_image_collection: PageImageCollection
    page_text_line_collection: PageTextLineCollection

    @property
    def height(self):
        return self.image.height

    @property
    def width(self):
        return self.image.width


class PageAssemblerStep(PipelineStep[PageAssemblerStepConfig, NoneTypePipelineStepResource]):

    def run(self, state: PipelineState, rnd: RandomState):
        page_layout = state.get_value(self.config.state_key_page_layout, PageLayout)
        background_image = state.get_value(self.config.state_key_background_image, Image)
        page_image_collection = state.get_value(
            self.config.state_key_page_image_collection,
            PageImageCollection,
        )
        page_text_line_collection = state.get_value(
            self.config.state_key_page_text_line_collection,
            PageTextLineCollection,
        )

        # Page background.
        assert background_image.mat.shape == (page_layout.height, page_layout.width, 3)
        assembled_image = background_image.copy()

        # Page images.
        for page_image in page_image_collection.page_images:
            page_image.box.fill_image(assembled_image, page_image.image)

        # Page text lines.
        for text_line in page_text_line_collection.text_lines:
            # Fill only the masked pixels.
            if text_line.score_map:
                text_line.score_map.fill_image(
                    assembled_image,
                    text_line.glyph_color,
                )
            else:
                text_line.mask.fill_image(
                    assembled_image,
                    text_line.image,
                )

        page = Page(
            image=assembled_image,
            page_image_collection=page_image_collection,
            page_text_line_collection=page_text_line_collection,
        )
        state.set_value(self.config.state_key_page, page)


page_assembler_step_factory = PipelineStepFactory(PageAssemblerStep)
