from typing import Optional, Sequence, Dict, Any, List

import attrs
from numpy.random import RandomState

from vkit.element import Image, Box
from vkit.pipeline.interface import (
    PipelineStep,
    NoneTypePipelineStepResource,
    PipelineStepFactory,
    PipelineState,
)
from vkit.engine.image import image_factory
from vkit.pipeline.step.page_layout import PageLayout


@attrs.define
class PageImageStepConifg:
    image_configs: Sequence[Dict[str, Any]]
    state_key_page_layout: str = 'page_layout'
    state_key_page_image_collection: str = 'page_image_collection'


@attrs.define
class PageImage:
    image: Image
    box: Box


@attrs.define
class PageImageCollection:
    height: int
    width: int
    page_images: Sequence[PageImage]


class PageImageStep(PipelineStep[PageImageStepConifg, NoneTypePipelineStepResource]):

    def __init__(
        self,
        config: PageImageStepConifg,
        resource: Optional[NoneTypePipelineStepResource] = None,
    ):
        super().__init__(config, resource)

        self.image_aggregator = image_factory.create(self.config.image_configs)

    def run(self, state: PipelineState, rnd: RandomState):
        page_layout = state.get_value(self.config.state_key_page_layout, PageLayout)

        page_images: List[PageImage] = []
        for layout_image in page_layout.layout_images:
            image = self.image_aggregator.run(
                {
                    'height': layout_image.box.height,
                    'width': layout_image.box.width,
                },
                rnd,
            )
            page_images.append(PageImage(
                image=image,
                box=layout_image.box,
            ))

        state.set_value(
            self.config.state_key_page_image_collection,
            PageImageCollection(
                height=page_layout.height,
                width=page_layout.width,
                page_images=page_images,
            ),
        )


page_image_step_factory = PipelineStepFactory(PageImageStep)
