from typing import Optional, Sequence, Dict, Set, List
from enum import Enum, unique

import attrs
from numpy.random import RandomState
import intervaltree

from vkit.utility import rnd_choice, normalize_to_keys_and_probs
from vkit.pipeline.interface import (
    PipelineStep,
    NoneTypePipelineStepResource,
    PipelineStepFactory,
    PipelineState,
)
from vkit.element import Box
from vkit.engine.font.type import FontEngineRunConfigGlyphSequence


@attrs.define
class LayoutImage:
    box: Box


@attrs.define
class LayoutTextLine:
    box: Box
    glyph_sequence: FontEngineRunConfigGlyphSequence


@attrs.define
class LayoutParagraph:
    box: Box
    layout_text_lines: Sequence[LayoutTextLine]


@attrs.define
class PageLayout:
    height: int
    width: int
    layout_images: Sequence[LayoutImage]
    layout_paragraphs: Sequence[LayoutParagraph]


@attrs.define
class LayoutSamplingConfig:
    prob_enable: Optional[float]
    num_layouts_min: int
    num_layouts_max: int
    empty_area_ratio_min: float
    height_ratio_min: float
    height_ratio_max: float
    width_ratio_min: float
    width_ratio_max: float


@attrs.define
class PageLayoutStepConfig:
    weight_random_text_line: float = 1
    weight_random_paragraph: float = 1

    image_sampling_config: LayoutSamplingConfig = attrs.field(
        factory=lambda: LayoutSamplingConfig(
            prob_enable=0.7,
            num_layouts_min=1,
            num_layouts_max=2,
            empty_area_ratio_min=0.2,
            height_ratio_min=0.158,
            height_ratio_max=0.316,
            width_ratio_min=0.158,
            width_ratio_max=0.316,
        )
    )

    paragraph_sampling_config: LayoutSamplingConfig = attrs.field(
        factory=lambda: LayoutSamplingConfig(
            prob_enable=None,
            num_layouts_min=3,
            num_layouts_max=10,
            empty_area_ratio_min=0.1,
            height_ratio_min=0.1,
            height_ratio_max=0.9,
            width_ratio_min=0.1,
            width_ratio_max=0.9,
        )
    )
    paragraph_text_line_gap_ratio_min: float = 0.0
    paragraph_text_line_gap_ratio_max: float = 0.75
    paragraph_text_line_width_ratio_min: float = 0.3
    paragraph_text_line_width_ratio_max: float = 1.0
    weight_paragraph_text_line_align_left: float = 1
    weight_paragraph_text_line_align_middle: float = 1
    weight_paragraph_text_line_align_right: float = 1

    text_line_sampling_config: LayoutSamplingConfig = attrs.field(
        factory=lambda: LayoutSamplingConfig(
            prob_enable=None,
            num_layouts_min=10,
            num_layouts_max=50,
            empty_area_ratio_min=0.1,
            height_ratio_min=0.01,
            height_ratio_max=0.1,
            width_ratio_min=0.03,
            width_ratio_max=0.9,
        )
    )

    state_key_height: str = 'height'
    state_key_width: str = 'width'
    state_key_page_layout: str = 'page_layout'


@unique
class PageLayoutStepKey(Enum):
    RANDOM_PARAGRAPH = 'random_paragraph'
    RANDOM_TEXT_LINE = 'random_text_line'

    PARAGRAPH_TEXT_LINE_ALIGN_LEFT = 'paragraph_text_line_align_left'
    PARAGRAPH_TEXT_LINE_ALIGN_MIDDLE = 'paragraph_text_line_align_middle'
    PARAGRAPH_TEXT_LINE_ALIGN_RIGHT = 'paragraph_text_line_align_right'


@attrs.define
class LayoutEmptyAreas:
    page_height: int
    page_width: int
    page_empty_area: int
    box_id_max: int
    id_to_box: Dict[int, Box]
    itv_tree_height: intervaltree.IntervalTree
    itv_tree_width: intervaltree.IntervalTree

    def add_box(self, box: Box):
        self.box_id_max += 1
        box_id = self.box_id_max

        self.id_to_box[box_id] = box
        self.itv_tree_height.addi(box.height, box.height + 1, box_id)
        self.itv_tree_width.addi(box.width, box.width + 1, box_id)
        self.page_empty_area += box.area

    def get_available_box_ids(self, height: int, width: int):
        height_box_ids: Set[int] = set(itv.data for itv in self.itv_tree_height[height:])
        width_box_ids: Set[int] = set(itv.data for itv in self.itv_tree_width[width:])
        return tuple(height_box_ids & width_box_ids)

    def sample_and_pop_empty_area(self, height: int, width: int, rnd: RandomState):
        box_ids = self.get_available_box_ids(height=height, width=width)
        if not box_ids:
            return None

        box_id = rnd_choice(rnd, box_ids)
        box = self.id_to_box.pop(box_id)
        self.itv_tree_height.removei(box.height, box.height + 1, box_id)
        self.itv_tree_width.removei(box.width, box.width + 1, box_id)
        self.page_empty_area -= box.area
        return box

    def get_empty_area_ratio(self):
        return self.page_empty_area / (self.page_height * self.page_width)


def initialize_layout_empty_areas(height: int, width: int):
    empty_areas = LayoutEmptyAreas(
        page_height=height,
        page_width=width,
        page_empty_area=0,
        box_id_max=-1,
        id_to_box={},
        itv_tree_height=intervaltree.IntervalTree(),
        itv_tree_width=intervaltree.IntervalTree(),
    )
    box = Box.from_shape(height, width)
    empty_areas.add_box(box)
    return empty_areas


def sample_and_fill_empty_area(
    empty_areas: LayoutEmptyAreas,
    height: int,
    width: int,
    rnd: RandomState,
):
    empty_box = empty_areas.sample_and_pop_empty_area(
        height=height,
        width=width,
        rnd=rnd,
    )
    if empty_box is None:
        return None

    up_offset_max = empty_box.height - height
    up_offset = rnd.randint(0, up_offset_max + 1)

    left_offset_max = empty_box.width - width
    left_offset = rnd.randint(0, left_offset_max + 1)

    up = empty_box.up + up_offset
    down = up + height - 1
    left = empty_box.left + left_offset
    right = left + width - 1

    filled_box = Box(up=up, down=down, left=left, right=right)

    if up_offset > 0:
        new_empty_box = attrs.evolve(empty_box, down=up - 1)
        empty_areas.add_box(new_empty_box)

    if down < empty_box.down:
        new_empty_box = attrs.evolve(empty_box, up=down + 1)
        empty_areas.add_box(new_empty_box)

    if empty_box.left < left:
        new_empty_box = Box(
            up=up,
            down=down,
            left=empty_box.left,
            right=left - 1,
        )
        empty_areas.add_box(new_empty_box)

    if right < empty_box.right:
        new_empty_box = Box(
            up=up,
            down=down,
            left=right + 1,
            right=empty_box.right,
        )
        empty_areas.add_box(new_empty_box)

    return filled_box


def sample_empty_area_based_on_config(
    config: LayoutSamplingConfig,
    empty_areas: LayoutEmptyAreas,
    rnd: RandomState,
    clip_width: bool = False,
):
    if config.prob_enable is not None and rnd.random() > config.prob_enable:
        return []

    height_min = round(config.height_ratio_min * empty_areas.page_height)
    height_max = round(config.height_ratio_max * empty_areas.page_height)
    width_min = round(config.width_ratio_min * empty_areas.page_height)
    width_max = round(config.width_ratio_max * empty_areas.page_height)

    num_layouts = rnd.randint(config.num_layouts_min, config.num_layouts_max + 1)
    boxes: List[Box] = []
    fail_cnt = 0

    while num_layouts > 0:
        if empty_areas.get_empty_area_ratio() < config.empty_area_ratio_min:
            break

        height = rnd.randint(height_min, height_max + 1)
        width = rnd.randint(width_min, width_max + 1)
        if clip_width:
            width = max(height, width)
        box = sample_and_fill_empty_area(
            empty_areas=empty_areas,
            height=height,
            width=width,
            rnd=rnd,
        )
        if box:
            boxes.append(box)
            num_layouts -= 1
            fail_cnt = 0
        else:
            fail_cnt += 1
            if fail_cnt >= 10:
                # Fast break.
                break

    return boxes


def sample_text_lines_under_paragraph(
    config: PageLayoutStepConfig,
    page_height: int,
    paragraph_tl_align_keys: Sequence[PageLayoutStepKey],
    paragraph_tl_align_probs: Sequence[float],
    box: Box,
    rnd: RandomState,
):
    height_min = round(config.text_line_sampling_config.height_ratio_min * page_height)
    height_max = round(config.text_line_sampling_config.height_ratio_max * page_height)

    text_line_height = rnd.randint(height_min, height_max + 1)
    text_line_height = min(text_line_height, box.height)

    text_line_width_min = round(config.paragraph_text_line_width_ratio_min * box.width)
    text_line_width_max = round(config.paragraph_text_line_width_ratio_max * box.width)

    gap_min = round(text_line_height * config.paragraph_text_line_gap_ratio_min)
    gap_max = round(text_line_height * config.paragraph_text_line_gap_ratio_max)
    gap = rnd.randint(gap_min, gap_max + 1)

    layout_text_lines: List[LayoutTextLine] = []
    hori_center = (box.left + box.right) // 2
    up = box.up
    while up < box.down:
        down = up + text_line_height - 1
        if down > box.down:
            break

        text_line_width = rnd.randint(text_line_width_min, text_line_width_max + 1)
        text_line_width = max(text_line_width, text_line_height)

        align_key = rnd_choice(rnd, paragraph_tl_align_keys, probs=paragraph_tl_align_probs)
        if align_key == PageLayoutStepKey.PARAGRAPH_TEXT_LINE_ALIGN_LEFT:
            left = box.left
        elif align_key == PageLayoutStepKey.PARAGRAPH_TEXT_LINE_ALIGN_MIDDLE:
            left = hori_center - text_line_width // 2 + 1
        elif align_key == PageLayoutStepKey.PARAGRAPH_TEXT_LINE_ALIGN_RIGHT:
            left = box.right - text_line_width + 1
        else:
            raise NotImplementedError()
        right = left + text_line_width - 1

        assert box.left <= left and right <= box.right
        assert box.up <= up and down <= box.down

        layout_text_lines.append(
            LayoutTextLine(
                box=Box(up=up, down=down, left=left, right=right),
                glyph_sequence=FontEngineRunConfigGlyphSequence.HORI_DEFAULT,
            )
        )

        up = down + gap

    assert layout_text_lines
    return layout_text_lines


class PageLayoutStep(PipelineStep[PageLayoutStepConfig, NoneTypePipelineStepResource]):

    def __init__(
        self,
        config: PageLayoutStepConfig,
        resource: Optional[NoneTypePipelineStepResource] = None,
    ):
        super().__init__(config, resource)

        self.mode_keys, self.mode_probs = normalize_to_keys_and_probs([
            (PageLayoutStepKey.RANDOM_PARAGRAPH, self.config.weight_random_paragraph),
            (PageLayoutStepKey.RANDOM_TEXT_LINE, self.config.weight_random_text_line)
        ])

        self.paragraph_tl_align_keys, self.paragraph_tl_align_probs = normalize_to_keys_and_probs([
            (
                PageLayoutStepKey.PARAGRAPH_TEXT_LINE_ALIGN_LEFT,
                self.config.weight_paragraph_text_line_align_left
            ),
            (
                PageLayoutStepKey.PARAGRAPH_TEXT_LINE_ALIGN_MIDDLE,
                self.config.weight_paragraph_text_line_align_middle
            ),
            (
                PageLayoutStepKey.PARAGRAPH_TEXT_LINE_ALIGN_RIGHT,
                self.config.weight_paragraph_text_line_align_right
            ),
        ])

    def generate_layout_elements(self, height: int, width: int, rnd: RandomState):
        empty_areas = initialize_layout_empty_areas(height, width)

        image_boxes = sample_empty_area_based_on_config(
            self.config.image_sampling_config,
            empty_areas,
            rnd,
        )
        layout_images = [LayoutImage(box) for box in image_boxes]

        mode_key = rnd_choice(rnd, self.mode_keys, probs=self.mode_probs)

        if mode_key == PageLayoutStepKey.RANDOM_PARAGRAPH:
            paragraph_boxes = sample_empty_area_based_on_config(
                self.config.paragraph_sampling_config,
                empty_areas,
                rnd,
                clip_width=True,
            )
            layout_paragraphs = [
                LayoutParagraph(
                    box=paragraph_box,
                    layout_text_lines=sample_text_lines_under_paragraph(
                        config=self.config,
                        page_height=height,
                        paragraph_tl_align_keys=self.paragraph_tl_align_keys,
                        paragraph_tl_align_probs=self.paragraph_tl_align_probs,
                        box=paragraph_box,
                        rnd=rnd,
                    )
                ) for paragraph_box in paragraph_boxes
            ]

        elif mode_key == PageLayoutStepKey.RANDOM_TEXT_LINE:
            text_line_boxes = sample_empty_area_based_on_config(
                self.config.text_line_sampling_config,
                empty_areas,
                rnd,
            )
            layout_paragraphs = [
                LayoutParagraph(
                    box=text_line_box,
                    layout_text_lines=[
                        LayoutTextLine(
                            box=text_line_box,
                            glyph_sequence=FontEngineRunConfigGlyphSequence.HORI_DEFAULT
                        )
                    ]
                ) for text_line_box in text_line_boxes
            ]

        else:
            raise NotImplementedError()

        return layout_images, layout_paragraphs

    def run(self, state: PipelineState, rnd: RandomState):
        height = state.get_value(self.config.state_key_height, int)
        width = state.get_value(self.config.state_key_width, int)

        while True:
            layout_images, layout_paragraphs = self.generate_layout_elements(
                height=height,
                width=width,
                rnd=rnd,
            )
            if layout_paragraphs:
                break

        page_layout = PageLayout(
            height=height,
            width=width,
            layout_images=layout_images,
            layout_paragraphs=layout_paragraphs,
        )
        state.set_value(self.config.state_key_page_layout, page_layout)


page_layout_step_factory = PipelineStepFactory(PageLayoutStep)
