from typing import Optional, Sequence, Dict, Any, List

import attrs
from numpy.random import RandomState

from vkit.element import LexiconCollection
from vkit.engine.font import (
    font_factory,
    FontEngineRunConfigStyle,
    FontCollection,
    TextLine,
)
from vkit.engine.char_sampler import char_sampler_factory
from vkit.engine.char_and_font_sampler import char_and_font_sampler_factory

from vkit.pipeline.interface import (
    PipelineStep,
    NoneTypePipelineStepResource,
    PipelineStepFactory,
    PipelineState,
)
from vkit.pipeline.step.page_layout import PageLayout


@attrs.define
class PageTextLineStepConfig:
    lexicon_collection_json: str
    font_collection_folder: str
    char_sampler_configs: Sequence[Dict[str, Any]]
    font_configs: Sequence[Dict[str, Any]]
    font_style: FontEngineRunConfigStyle = attrs.field(factory=FontEngineRunConfigStyle)
    return_font_variant: bool = False
    state_key_page_layout: str = 'page_layout'
    state_key_page_text_line_collection: str = 'page_text_line_collection'


@attrs.define
class PageTextLineCollection:
    height: int
    width: int
    text_lines: Sequence[TextLine]


class PageTextLineStep(PipelineStep[PageTextLineStepConfig, NoneTypePipelineStepResource]):

    def __init__(
        self,
        config: PageTextLineStepConfig,
        resource: Optional[NoneTypePipelineStepResource] = None,
    ):
        super().__init__(config, resource)

        lexicon_collection = LexiconCollection.from_file(self.config.lexicon_collection_json)
        font_collection = FontCollection.from_folder(self.config.font_collection_folder)
        char_sampler_aggregator = char_sampler_factory.create(
            self.config.char_sampler_configs,
            {
                'lexicon_collection': lexicon_collection,
            },
        )
        self.char_and_font_sampler = char_and_font_sampler_factory.create(
            {},
            {
                'lexicon_collection': lexicon_collection,
                'font_collection': font_collection,
                'char_sampler_aggregator': char_sampler_aggregator,
            },
        )
        self.font_aggregator = font_factory.create(self.config.font_configs)

    def run(self, state: PipelineState, rnd: RandomState):
        page_layout = state.get_value(self.config.state_key_page_layout, PageLayout)

        text_lines: List[TextLine] = []
        for layout_paragraph in page_layout.layout_paragraphs:
            for layout_text_line in layout_paragraph.layout_text_lines:

                while True:
                    char_and_font = self.char_and_font_sampler.run(
                        config={
                            'height': layout_text_line.box.height,
                            'width': layout_text_line.box.width,
                        },
                        rnd=rnd,
                    )
                    if char_and_font:
                        break

                text_line = self.font_aggregator.run(
                    config={
                        'height': layout_text_line.box.height,
                        'width': layout_text_line.box.width,
                        'chars': char_and_font.chars,
                        'font_variant': char_and_font.font_variant,
                        'glyph_sequence': layout_text_line.glyph_sequence,
                        'style': self.config.font_style,
                        'return_font_variant': self.config.return_font_variant,
                    },
                    rnd=rnd,
                )
                if text_line:
                    text_line = text_line.to_shifted_text_line(
                        y_offset=layout_text_line.box.up,
                        x_offset=layout_text_line.box.left,
                    )
                    text_lines.extend(text_line.split())

        assert text_lines
        state.set_value(
            self.config.state_key_page_text_line_collection,
            PageTextLineCollection(
                height=page_layout.height,
                width=page_layout.width,
                text_lines=text_lines,
            )
        )


page_text_line_step_factory = PipelineStepFactory(PageTextLineStep)
