from typing import List, Sequence

import attrs
from numpy.random import RandomState

from vkit.element import Point, Box, Mask, ScoreMap
from vkit.pipeline.interface import (
    PipelineStep,
    NoneTypePipelineStepResource,
    PipelineStepFactory,
    PipelineState,
)
from vkit.pipeline.step.page_assembler import Page


@attrs.define
class PageTextLineLabelStepConfig:
    enable_font_size_score_map: bool = True
    enable_boundary_mask: bool = True
    boundary_dilate_ratio: float = 0.5
    enable_boundary_score_map: bool = False

    # Keys.
    state_key_page: str = 'page'
    state_key_page_text_line_mask = 'page_text_line_mask'
    state_key_page_text_line_font_size_score_map = 'page_text_line_font_size_score_map'
    state_key_page_text_line_boundary_mask = 'page_text_line_boundary_mask'
    state_key_page_text_line_and_boundary_mask = 'page_text_line_and_boundary_mask'
    state_key_page_text_line_boundary_score_map = 'page_text_line_boundary_score_map'


class PageTextLineLabelStep(
    PipelineStep[PageTextLineLabelStepConfig, NoneTypePipelineStepResource]
):

    def generate_page_text_line_mask(self, page: Page):
        page_text_line_mask = Mask.from_shapable(page)

        text_lines = page.page_text_line_collection.text_lines
        for text_line in text_lines:
            text_line.box.fill_mask(page_text_line_mask)
        return page_text_line_mask

    def generate_page_text_line_font_size_score_map(self, page: Page):
        page_text_line_font_size_score_map = ScoreMap.from_shapable(
            page,
            score_as_prob=False,
        )

        text_lines = page.page_text_line_collection.text_lines
        for text_line in text_lines:
            text_line.box.fill_score_map(
                page_text_line_font_size_score_map,
                text_line.font_size,
            )
        return page_text_line_font_size_score_map

    def generate_text_line_boxes_and_dilated_boxes(self, page: Page):
        text_lines = page.page_text_line_collection.text_lines
        text_lines = sorted(text_lines, key=lambda tl: tl.font_size, reverse=True)

        boxes: List[Box] = []
        dilated_boxes: List[Box] = []

        for text_line in text_lines:
            box = text_line.box
            boxes.append(box)

            dilated_box = box.to_dilated_box(self.config.boundary_dilate_ratio, clip_long_side=True)
            dilated_box = dilated_box.to_clipped_box(page)
            dilated_boxes.append(dilated_box)

        return boxes, dilated_boxes

    @staticmethod
    def generate_dilated_only_boxes(
        box: Box,
        dilated_box: Box,
    ):
        dilated_up_box = attrs.evolve(
            dilated_box,
            down=box.up - 1,
        )
        if dilated_up_box.up > dilated_box.down:
            dilated_up_box = None

        dilated_down_box = attrs.evolve(
            dilated_box,
            up=box.down + 1,
        )
        if dilated_down_box.up > dilated_down_box.down:
            dilated_down_box = None

        dilated_left_box = attrs.evolve(
            box,
            left=dilated_box.left,
            right=box.left - 1,
        )
        if dilated_left_box.left > dilated_left_box.right:
            dilated_left_box = None

        dilated_right_box = attrs.evolve(
            box,
            left=box.right + 1,
            right=dilated_box.right,
        )
        if dilated_right_box.left > dilated_right_box.right:
            dilated_right_box = None

        return (
            dilated_up_box,
            dilated_down_box,
            dilated_left_box,
            dilated_right_box,
        )

    def generate_page_text_line_boundary_masks(
        self,
        page: Page,
        boxes: Sequence[Box],
        dilated_boxes: Sequence[Box],
        page_text_line_mask: Mask,
    ):
        boundary_mask = Mask.from_shapable(page)

        for box, dilated_box in zip(boxes, dilated_boxes):
            dilated_only_boxes = PageTextLineLabelStep.generate_dilated_only_boxes(box, dilated_box)
            for dilated_only_box in dilated_only_boxes:
                if dilated_only_box:
                    dilated_only_box.fill_mask(boundary_mask)

        page_text_line_mask.fill_mask(boundary_mask, 0)

        text_line_and_boundary_mask = boundary_mask.copy()
        page_text_line_mask.fill_mask(text_line_and_boundary_mask)

        return boundary_mask, text_line_and_boundary_mask

    def generate_page_text_line_boundary_score_map(
        self,
        page: Page,
        boxes: Sequence[Box],
        dilated_boxes: Sequence[Box],
        page_text_line_boundary_mask: Mask,
    ):
        boundary_score_map = ScoreMap.from_shapable(page, fill_value=1.0)

        for box, dilated_box in zip(boxes, dilated_boxes):
            (
                dilated_up_box,
                dilated_down_box,
                dilated_left_box,
                dilated_right_box,
            ) = PageTextLineLabelStep.generate_dilated_only_boxes(box, dilated_box)

            if dilated_up_box:
                boundary_score_map.fill_by_quad_interpolation(
                    point0=Point(y=box.up, x=box.right),
                    point1=Point(y=box.up, x=box.left),
                    point2=Point(y=dilated_box.up, x=dilated_box.left),
                    point3=Point(y=dilated_box.up, x=dilated_box.right),
                    func_np_uv_to_mat=lambda np_uv: np_uv[:, :, 1],
                    keep_min_value=True,
                )

            if dilated_down_box:
                boundary_score_map.fill_by_quad_interpolation(
                    point0=Point(y=box.down, x=box.left),
                    point1=Point(y=box.down, x=box.right),
                    point2=Point(y=dilated_box.down, x=dilated_box.right),
                    point3=Point(y=dilated_box.down, x=dilated_box.left),
                    func_np_uv_to_mat=lambda np_uv: np_uv[:, :, 1],
                    keep_min_value=True,
                )

            if dilated_left_box:
                boundary_score_map.fill_by_quad_interpolation(
                    point0=Point(y=box.up, x=box.left),
                    point1=Point(y=box.down, x=box.left),
                    point2=Point(y=dilated_box.down, x=dilated_box.left),
                    point3=Point(y=dilated_box.up, x=dilated_box.left),
                    func_np_uv_to_mat=lambda np_uv: np_uv[:, :, 1],
                    keep_min_value=True,
                )

            if dilated_right_box:
                boundary_score_map.fill_by_quad_interpolation(
                    point0=Point(y=box.down, x=box.right),
                    point1=Point(y=box.up, x=box.right),
                    point2=Point(y=dilated_box.up, x=dilated_box.right),
                    point3=Point(y=dilated_box.down, x=dilated_box.right),
                    func_np_uv_to_mat=lambda np_uv: np_uv[:, :, 1],
                    keep_min_value=True,
                )

        page_text_line_boundary_mask.to_inverted_mask().fill_score_map(boundary_score_map, 0.0)
        return boundary_score_map

    def run(self, state: PipelineState, rnd: RandomState):
        page = state.get_value(self.config.state_key_page, Page)

        page_text_line_mask = self.generate_page_text_line_mask(page)
        state.set_value(self.config.state_key_page_text_line_mask, page_text_line_mask)

        if self.config.enable_font_size_score_map:
            state.set_value(
                self.config.state_key_page_text_line_font_size_score_map,
                self.generate_page_text_line_font_size_score_map(page),
            )

        boxes, dilated_boxes = self.generate_text_line_boxes_and_dilated_boxes(page)
        if self.config.enable_boundary_mask:
            (
                page_text_line_boundary_mask,
                page_text_line_and_boundary_mask,
            ) = self.generate_page_text_line_boundary_masks(
                page,
                boxes,
                dilated_boxes,
                page_text_line_mask,
            )
            state.set_value(
                self.config.state_key_page_text_line_boundary_mask,
                page_text_line_boundary_mask,
            )
            state.set_value(
                self.config.state_key_page_text_line_and_boundary_mask,
                page_text_line_and_boundary_mask,
            )

            if self.config.enable_boundary_score_map:
                page_text_line_boundary_score_map = self.generate_page_text_line_boundary_score_map(
                    page,
                    boxes,
                    dilated_boxes,
                    page_text_line_boundary_mask,
                )
                state.set_value(
                    self.config.state_key_page_text_line_boundary_score_map,
                    page_text_line_boundary_score_map,
                )


page_text_line_label_step = PipelineStepFactory(PageTextLineLabelStep)
