from typing import Sequence, Optional, TypeVar, Any, Type, Union, Tuple, Dict, List
import subprocess
from os import PathLike
from collections import abc

from numpy.random import RandomState
import iolite as io
import cattrs
from cattrs.errors import ClassValidationError

from vkit.utility import PathType

T_ITEM = TypeVar('T_ITEM')


def is_path_type(path: PathType):
    return isinstance(path, (str, bytes, PathLike))  # type: ignore


def get_data_folder(file: PathType):
    proc = subprocess.run(
        f'pyproject-data-folder "$VKIT_ROOT" "$VKIT_DATA" "{file}"',
        shell=True,
        capture_output=True,
        text=True,
    )
    assert proc.returncode == 0

    data_folder = proc.stdout.strip()
    assert data_folder

    io.folder(data_folder, touch=True)

    return data_folder


def rnd_choice(
    rnd: RandomState,
    items: Sequence[T_ITEM],
    probs: Optional[Sequence[float]] = None,
) -> T_ITEM:
    idx = rnd.choice(len(items), p=probs)
    return items[idx]


T_TGT = TypeVar('T_TGT')


def dyn_structure(
    dyn_object: Any,
    target_cls: Type[T_TGT],
    support_path_type: bool = False,
    support_none_type: bool = False,
) -> T_TGT:
    if support_none_type and dyn_object is None:
        return target_cls()

    if support_path_type and is_path_type(dyn_object):
        dyn_object = io.read_json(dyn_object, expandvars=True)

    if isinstance(dyn_object, abc.Mapping):
        try:
            return cattrs.structure(dyn_object, target_cls)
        except ClassValidationError:
            return target_cls(**dyn_object)

    if isinstance(dyn_object, target_cls):
        return dyn_object

    raise NotImplementedError()


T_KEY = TypeVar('T_KEY')


def normalize_to_keys_and_probs(
    key_weight_items: Union[Sequence[Tuple[T_KEY, float]], Dict[T_KEY, float]]
) -> Tuple[Sequence[T_KEY], Sequence[float]]:
    keys: List[T_KEY] = []
    weights: List[float] = []

    if isinstance(key_weight_items, abc.Sequence):
        for key, weight in key_weight_items:
            keys.append(key)
            weights.append(weight)
    elif isinstance(key_weight_items, abc.Mapping):  # type: ignore
        for key, weight in key_weight_items.items():
            keys.append(key)
            weights.append(weight)
    else:
        raise NotImplementedError()

    total = sum(weights)
    probs = [weight / total for weight in weights]
    return keys, probs
