# flake8: noqa
from typing import Sequence
from xml.etree import ElementTree
from pathlib import Path
import logging
import concurrent.futures
import shutil

import cv2 as cv
import numpy as np
import iolite as io
from tqdm import tqdm

from vkit.utility import PathType
from vkit.element import Image, ImageKind, Polygon, Mask
from vkit.element.visualization import visualize_polygons, visualize_mask
from vkit.augmentation.photometric_distortion import (
    boundary_equalization,
    BoundaryEqualizationConfig,
)

logger = logging.getLogger(__name__)


def xml_node_to_tree(node: ElementTree.Element):
    data = {}
    for child_node in node:
        value = ''
        if child_node.text:
            value = child_node.text.strip()
        if value:
            data[child_node.tag] = value
        else:
            data[child_node.tag] = xml_node_to_tree(child_node)
    return data


def parse_label_xml(xml_file: PathType):
    root = ElementTree.parse(xml_file).getroot()
    filename = None
    size = None
    objects = []
    for child_node in root:
        if child_node.tag == 'filename':
            assert child_node.text
            filename = child_node.text.strip()
        elif child_node.tag == 'size':
            size = xml_node_to_tree(child_node)
        elif child_node.tag == 'object':
            objects.append(xml_node_to_tree(child_node))
        else:
            # print('ignore', child_node)
            pass

    return {
        'filename': filename,
        'size': size,
        'objects': objects,
    }


def get_shape_and_polygons_from_label_xml(xml_file: PathType):
    data = parse_label_xml(xml_file)

    height = int(data['size']['height'])
    width = int(data['size']['width'])

    polygons = []
    for obj in data['objects']:
        bndbox = obj['bndbox']
        xy_pairs = [
            (bndbox['x1'], bndbox['y1']),
            (bndbox['x2'], bndbox['y2']),
            (bndbox['x3'], bndbox['y3']),
            (bndbox['x4'], bndbox['y4']),
        ]
        polygon = Polygon.from_xy_pairs(xy_pairs)
        polygon = polygon.to_shrank_polygon(0.8)
        polygon = polygon.to_bounding_rectangular_polygon()
        polygons.append(polygon)

    return (height, width), polygons


def get_canny_mask(image: Image, polygons: Sequence[Polygon]):
    image = boundary_equalization.distort_image(BoundaryEqualizationConfig(), image=image)
    grayscale_image = image.to_grayscale_image()
    pixels = []
    for polygon in polygons:
        mat, _ = polygon.extract_np_array(grayscale_image.mat)
        pixels.append(mat.reshape(-1))
    pixels = np.concatenate(pixels)
    mat_mean = np.mean(pixels)
    mat_range = np.max(pixels) - np.min(pixels)

    sigma = 0.3
    low = round(max(0, mat_mean - mat_range * sigma))
    high = round(min(255, mat_mean + mat_range * sigma))
    np_mask = cv.Canny(grayscale_image.mat, low, high)
    return Mask(np_mask)


def check_if_polygon_is_dirty(image: Image, polygon: Polygon):
    # Test 0.
    # mat, np_mask = polygon.extract_np_array(image.mat)
    # rgb_min = np.amin(mat[np_mask], axis=0)
    # grayscale_min = np.dot(rgb_min, (0.299, 0.587, 0.114))
    # rgb_max = np.amax(mat[np_mask], axis=0)
    # grayscale_max = np.dot(rgb_max, (0.299, 0.587, 0.114))
    # if grayscale_max - grayscale_min > 50:
    #     return True
    assert image.kind == ImageKind.GRAYSCALE
    mat, np_mask = polygon.extract_np_array(image.mat)
    # mat_mean = np.mean(mat[np_mask])
    # mat_range = np.max(mat[np_mask]) - np.min(mat[np_mask])
    # sigma = 0.1
    # low = round(max(0, mat_mean - mat_range * sigma))
    # high = round(min(255, mat_mean + mat_range * sigma))
    # blur = cv.GaussianBlur(mat[np_mask], (3, 3), 0)
    otsu_thr, _ = cv.threshold(mat[np_mask], 0, 255, cv.THRESH_BINARY + cv.THRESH_OTSU)
    mat = cv.Canny(mat, 0.5 * otsu_thr, otsu_thr)

    # mat, np_mask = polygon.extract_np_array(mask.mat)

    # Test 1.
    count = (mat[np_mask] > 0).sum()
    ratio = count / max(np_mask.shape)
    if count < 10:
        return False
    if ratio > 0.2:
        return True

    # Test 2. (text in one side)
    height, width = np_mask.shape
    if height > width:
        half_height = height // 2
        count_a = (mat[:half_height][np_mask[:half_height] > 0] > 0).sum()
        count_b = (mat[half_height:][np_mask[half_height:] > 0] > 0).sum()
    else:
        half_width = width // 2
        count_a = (mat[:, :half_width][np_mask[:, :half_width] > 0] > 0).sum()
        count_b = (mat[:, half_width:][np_mask[:, half_width:] > 0] > 0).sum()

    ratio = min(count_a, count_b) / max(count_a, count_b)
    if ratio < 0.2:
        return True

    # Test 3. (text in center, or in two sides.)
    if height > width:
        one_fourth_height = height // 4
        half_height = height // 2
        count_a = (mat[:one_fourth_height][np_mask[:one_fourth_height] > 0] > 0).sum()
        count_b = (
            mat[one_fourth_height:one_fourth_height
                + half_height][np_mask[one_fourth_height:one_fourth_height + half_height] > 0] > 0
        ).sum()
        count_c = (
            mat[one_fourth_height + half_height:][np_mask[one_fourth_height + half_height:] > 0] > 0
        ).sum()

    else:
        one_fourth_width = width // 4
        half_width = width // 2
        count_a = (mat[:, :one_fourth_width][np_mask[:, :one_fourth_width] > 0] > 0).sum()
        count_b = (
            mat[:, one_fourth_width:one_fourth_width
                + half_width][np_mask[:, one_fourth_width:one_fourth_width + half_width] > 0] > 0
        ).sum()
        count_c = (
            mat[:, one_fourth_width + half_width:][np_mask[:, one_fourth_width + half_width:] > 0] >
            0
        ).sum()

    count_mid = count_b
    count_non_mid = count_a + count_c
    ratio = min(count_mid, count_non_mid) / max(count_mid, count_non_mid)
    if ratio < 0.2:
        return True

    return False


def extarct_and_transform_polygon(image: Image, polygon: Polygon):
    internals = polygon.to_fill_np_array_internals()
    mat = internals.bounding_box.extract_np_array(image.mat)
    np_mask = internals.get_np_mask()
    shifted_polygon = internals.get_shifted_polygon()

    height, width = np_mask.shape
    count = (np_mask > 0).sum()
    ratio = count / (height * width)
    new_height = round(height * ratio)
    new_width = round(width * ratio)

    dst_xy_pairs = [
        (0, 0),
        (new_width - 1, 0),
        (new_width - 1, new_height - 1),
        (0, new_height - 1),
    ]
    trans_mat = cv.getPerspectiveTransform(
        np.array(shifted_polygon.to_xy_pairs(), dtype=np.float32),
        np.array(dst_xy_pairs, dtype=np.float32),
        cv.DECOMP_SVD,
    )
    mat = cv.warpPerspective(mat, trans_mat, (new_width, new_height))

    if np.unique(mat.reshape(-1, 3), axis=0).shape[0] < 500 \
            or np.mean(np.std(mat.reshape(-1, 3), axis=0)) < 3:
        # Not enough color or not vary enough.
        return None
    else:
        return Image(mat=mat)


def extract_and_save_clean_background(
    xml_fd: Path,
    image_fd: Path,
    out_fd: Path,
    name: str,
):
    xml_file = io.file(xml_fd / f'{name}..xml', exists=True)
    image_file = io.file(image_fd / f'{name}.jpg', exists=True)

    image = Image.from_file(image_file)
    if image.kind not in (ImageKind.RGB, ImageKind.RGBA):
        return
    image = image.to_rgb_image()

    _, polygons = get_shape_and_polygons_from_label_xml(xml_file)
    # canny_mask = get_canny_mask(image, polygons)

    for_test_image = image.to_grayscale_image()
    polygons_clean = []
    polygons_dirty = []
    for polygon in polygons:
        if check_if_polygon_is_dirty(for_test_image, polygon):
            polygons_dirty.append(polygon)
        else:
            polygons_clean.append(polygon)

    if not polygons_clean:
        return

    polygon_images = []
    for polygon in polygons_clean:
        polygon_image = extarct_and_transform_polygon(image, polygon)
        if polygon_image:
            polygon_images.append(polygon_image)

    if not polygon_images:
        return

    sub_out_fd = io.folder(out_fd / name, touch=True)
    assert polygon_images
    for idx, polygon_image in enumerate(polygon_images):
        polygon_image.to_file(sub_out_fd / f'{idx}.png')


def proc_extract_and_save_clean_background(args: Sequence[str]):
    (xml_folder, image_folder, output_folder, name) = args

    xml_fd = io.folder(xml_folder, exists=True)
    image_fd = io.folder(image_folder, exists=True)
    out_fd = io.folder(output_folder, exists=True)
    try:
        extract_and_save_clean_background(
            xml_fd=xml_fd,
            image_fd=image_fd,
            out_fd=out_fd,
            name=name,
        )
    except Exception:
        logger.exception(f'failed {name}')


def extract_all(
    xml_fd: Path,
    image_fd: Path,
    out_fd: Path,
    skip_exists: bool = False,
):
    names = [xml_file.name.split('..')[0] for xml_file in xml_fd.glob('*.xml')]
    names = [name for name in names if 'table_spider' not in name and 'taobao' not in name]
    if skip_exists:
        new_names = []
        for name in names:
            if (out_fd / name).exists():
                continue
            new_names.append(name)
        names = new_names

    print(len(names), 'names')
    with concurrent.futures.ProcessPoolExecutor(max_workers=8) as executor:
        for _ in tqdm(
            executor.map(
                proc_extract_and_save_clean_background,
                [(str(xml_fd), str(image_fd), str(out_fd), name) for name in names]
            ),
            total=len(names),
        ):
            pass


def get_rgb_std(image: Image):
    # Per-channel std.
    rgb_std = np.std(image.mat.reshape(-1, 3), axis=0)
    return float(np.mean(rgb_std))


def sort_image_by_area():
    in_fd = io.folder(
        '/Users/huntzhan/Project/personal/vkit-x/vkit-data/vkit_sln/dataset_wtw/background',
        exists=True,
    )
    out_fd = io.folder(
        '/Users/huntzhan/Project/personal/vkit-x/vkit-data/vkit_sln/dataset_wtw/background2',
        reset=True,
    )
    image_file_area_pairs = []
    for image_file in in_fd.glob('**/*.png'):
        image = Image.from_file(image_file)
        area = image.height * image.width
        image_file_area_pairs.append((image_file, area))
    image_file_area_pairs = sorted(image_file_area_pairs, key=lambda p: p[1], reverse=True)
    for idx, (image_file, _) in enumerate(image_file_area_pairs):
        shutil.copyfile(image_file, out_fd / f'{idx}.png')


def debug():
    name = 'car-invoice-img02508'
    wtw_train_folder = '/Users/huntzhan/Project/personal/vkit-x/vkit-data/vkit_sln/dataset_wtw/wtw-train'

    image = Image.from_file(f'{wtw_train_folder}/images/{name}.jpg')

    debug_fd = '/Users/huntzhan/Project/personal/vkit-x/vkit-data/vkit_sln/dataset_wtw'

    boundary_equalization.distort_image(BoundaryEqualizationConfig(),
                                        image=image).to_file(f'{debug_fd}/debug-aug.png')

    (height, width
     ), polygons = get_shape_and_polygons_from_label_xml(f'{wtw_train_folder}/xml/{name}..xml')
    assert image.shape == (height, width)
    visualize_polygons(image, polygons).to_file(f'{debug_fd}/debug.png')

    canny_mask = get_canny_mask(image, polygons)
    visualize_polygons(visualize_mask(canny_mask), polygons).to_file(f'{debug_fd}/debug-edged.png')

    for_test_image = image.to_grayscale_image()
    polygons_clean = []
    polygons_dirty = []
    for polygon in polygons:
        if check_if_polygon_is_dirty(for_test_image, polygon):
            polygons_dirty.append(polygon)
        else:
            polygons_clean.append(polygon)

    visualize_polygons(image, polygons_clean, fill='green').to_file(f'{debug_fd}/debug-clean.png')
    visualize_polygons(image, polygons_dirty, fill='red').to_file(f'{debug_fd}/debug-dirty.png')

    for idx, polygon in enumerate(polygons_clean):
        polygon_image = extarct_and_transform_polygon(image, polygon)
        if polygon_image:
            polygon_image.to_file(f'{debug_fd}/debug-box/{idx}.png')


def debug2():
    wtw_train_folder = '/Users/huntzhan/Project/personal/vkit-x/vkit-data/vkit_sln/dataset_wtw/wtw-train'
    extract_all(
        xml_fd=io.folder(f'{wtw_train_folder}/xml/', exists=True),
        image_fd=io.folder(f'{wtw_train_folder}/images/', exists=True),
        out_fd=io.folder(
            '/Users/huntzhan/Project/personal/vkit-x/vkit-data/vkit_sln/dataset_wtw/background/',
            touch=True
        ),
        skip_exists=True,
    )


def debug3():
    background_fd = io.folder(
        '/Users/huntzhan/Project/personal/vkit-x/vkit-data/vkit_sln/dataset_wtw/background/',
        exists=True,
    )
    file_std_pairs = []
    for image_file in tqdm(list(background_fd.glob('**/*.png'))):
        image = Image.from_file(image_file)
        std = get_rgb_std(image)
        file_std_pairs.append((image_file, std))

    file_std_pairs = sorted(file_std_pairs, key=lambda p: p[1], reverse=True)
    np.median([std for _, std in file_std_pairs])
    new_file_std_pairs = [(file, std) for file, std in file_std_pairs if std >= 6.2]
    breakpoint()


def debug4():
    synthesizer_folder = '/Users/huntzhan/Project/personal/vkit-x/vkit-data/vkit_sln/dataset_wtw/synthesizer'
    image_fd = io.folder(
        f'{synthesizer_folder}/image',
        exists=True,
    )
    metas = []
    for image_file in image_fd.glob('**/*.png'):
        if 'table_spider' in image_file.name:
            continue
        if 'taobao' in image_file.name:
            continue

        image = Image.from_file(image_file)
        image = image.to_grayscale_image()

        metas.append({
            'image_file': str(image_file.relative_to(image_fd)),
            'grayscale_mean': np.mean(image.mat),
            'grayscale_std': np.std(image.mat),
        })
    io.write_json(
        f'{synthesizer_folder}/metas.json',
        metas,
        indent=2,
        ensure_ascii=True,
    )
