from pathlib import Path
from typing import Callable, DefaultDict, Tuple, List, Set
import logging
from collections import defaultdict
import hashlib
import shutil
import multiprocessing

import iolite as io
from tqdm import tqdm
from PIL import ImageFont as PILImageFont
from fontTools.ttLib import TTFont, TTLibError
import freetype
import numpy as np

from vkit.element import LexiconCollection
from vkit.engine.font.type import (
    FontMeta,
    FontKind,
    FontCollectionFolderTree,
    FontCollection,
)
from vkit.engine.font.freetype import (
    trim_char_np_image_vert,
    trim_char_np_image_hori,
)
from .tex import TexRenderBasic

logger = logging.getLogger(__name__)


def calculate_hash(path: Path):
    sha256_algo = hashlib.sha256()
    with open(path, 'rb') as fin:
        # 64KB block.
        for block in iter(lambda: fin.read(65536), b''):
            sha256_algo.update(block)
    return sha256_algo.hexdigest()


def search_font_files(
    font_folders: str,
    truetype: bool = True,
    opentype: bool = True,
):
    exts = []
    if truetype:
        exts.extend(['ttf', 'ttc'])
    if opentype:
        # OTC is ignored since it's not commonly used.
        exts.extend(['otf'])

    for ext in list(exts):
        exts.append(ext.upper())

    # Remove duplicated file.
    hash_to_paths: DefaultDict[str, List[Path]] = defaultdict(list)
    for font_folder in font_folders.split(','):
        input_fd = io.folder(font_folder.strip(), exists=True)
        for ext in exts:
            for font_path in input_fd.glob(f'**/*.{ext}'):
                hash_to_paths[calculate_hash(font_path)].append(font_path)

    filename_to_paths: DefaultDict[str, List[Path]] = defaultdict(list)
    for paths_with_same_hash in hash_to_paths.values():
        if len(paths_with_same_hash) > 1:
            logger.info(f'paths with same hash: {paths_with_same_hash}')
        path = paths_with_same_hash[0]
        filename_to_paths[path.name].append(path)

    for _, paths in filename_to_paths.items():
        if len(paths) == 1:
            yield paths[0], paths[0].name
            continue

        # Resolue name conflict.
        logger.info(f'resolve conflict: {paths}')
        for idx, path in enumerate(paths):
            prefix, ext = path.name.rsplit('.', maxsplit=1)
            yield path, f'{prefix}-{idx}.{ext}'


def flatten_font_files(font_folders: str, output: str):
    assert ' ' not in output
    output_fd = io.folder(output, touch=True)
    for font_path, flatten_name in search_font_files(font_folders):
        # Remove space in filename.
        flatten_name = '-'.join(flatten_name.split())
        shutil.copyfile(font_path, output_fd / flatten_name)


def classify_chinese_font(args: Tuple[str, str]):
    font_file, render_folder = args

    text_sim = '永我能吞下玻璃而不伤身体'
    text_tra = '永我能吞下玻璃而不傷身體'

    render_basic = TexRenderBasic(render_folder)

    render_basic.generate_tex_file(
        text=text_sim,
        raw_font_file=font_file,
        ttc_font_index=None,
        font_size=10,
        letter_space=0.0,
        word_space=1.0,
        punctuation_space=0.0,
        rgb_hex='000000',
        tag='default',
    )

    runtime_error = False
    try:
        render_basic.compile_tex_file(tag='default')
    except RuntimeError:
        runtime_error = True
    if runtime_error:
        return font_file, False, False, True

    support_sim = not render_basic.get_not_supported_chars(tag='default')

    render_basic.generate_tex_file(
        text=text_tra,
        raw_font_file=font_file,
        ttc_font_index=None,
        font_size=10,
        letter_space=0.0,
        word_space=1.0,
        punctuation_space=0.0,
        rgb_hex='000000',
        tag='default',
    )

    runtime_error = False
    try:
        render_basic.compile_tex_file(tag='default')
    except RuntimeError:
        runtime_error = True
    if runtime_error:
        return font_file, False, False, True

    support_tra = not render_basic.get_not_supported_chars(tag='default')

    return font_file, support_sim, support_tra, False


def classify_chinese_fonts(font_folder: str, render_folder: str, output: str):
    font_files = list(font_file for font_file, _ in search_font_files(font_folder))
    results = []
    with multiprocessing.Pool() as pool:
        for font_file, support_sim, support_tra, runtime_error in tqdm(
            pool.imap_unordered(
                classify_chinese_font,
                [(str(font_file), render_folder) for font_file in font_files],
                chunksize=1,
            )
        ):
            results.append({
                'font_file': str(font_file),
                'support_sim': support_sim,
                'support_tra': support_tra,
                'runtime_error': runtime_error,
            })

    io.write_json(io.file(output), results, indent=2)


def select_chinese_fonts(cla_results: str, output: str):
    resutls = io.read_json(io.file(cla_results, exists=True))

    assert ' ' not in output
    output_fd = io.folder(output, touch=True)
    for result in resutls:
        if result['support_sim'] or result['support_tra']:
            font_file = result['font_file']

            in_blacklist = False
            for keyword in (
                'ヒ',
                'LastResort',
                'Wingdings',
                'CourierOblique',
            ):
                if keyword in font_file:
                    in_blacklist = True
                    break

            if in_blacklist:
                logger.info(f'In blacklist {font_file}')
                continue

            logger.info(f'Select {font_file}')
            filename = font_file.split('/')[-1]
            shutil.copyfile(font_file, output_fd / filename)


def get_font_index_max(path: str):
    index0 = 0
    while True:
        try:
            PILImageFont.truetype(
                font=path,
                size=50,
                index=index0,
            )
            index0 += 1
        except OSError:
            break

    index1 = 0
    while True:
        try:
            TTFont(path, fontNumber=index1)
            index1 += 1
        except TTLibError:
            break

    # Inclusive.
    return min(index0, index1) - 1


def check_no_collection_in_ttf(font_folder: str):
    ttf_files = []
    ttc_files = []
    for font_file, _ in search_font_files(font_folder, truetype=True, opentype=False):
        if str(font_file).lower().endswith('ttf'):
            ttf_files.append(font_file)
        else:
            ttc_files.append(font_file)

    logger.info('TTF')
    for font_file in ttf_files:
        font_file = str(font_file)
        logger.info(font_file)
        index_max = get_font_index_max(font_file)
        logger.info(index_max)
        assert index_max == 1

    logger.info('=' * 20)

    logger.info('TTC')
    for font_file in ttc_files:
        font_file = str(font_file)
        logger.info(font_file)
        index_max = get_font_index_max(font_file)
        logger.info(index_max)
        assert index_max > 1


def proc_find_xelatex_supported_chars(args: Tuple[str, List[str], str]):
    font_file, chars, render_folder = args

    not_supported_chars: Set[str] = set()
    chunksize = 100
    begin = 0

    render_basic = TexRenderBasic(render_folder)

    try:
        while begin < len(chars):
            end = begin + chunksize
            logger.info(f'{font_file}, [{begin}, {end}]')

            text = ''.join(chars[begin:end])
            render_basic.generate_tex_file(
                text=text,
                raw_font_file=font_file,
                # NOTE: we assume TTC support the same set of chars in every font face.
                ttc_font_index=None,
                font_size=10,
                letter_space=0.0,
                word_space=1.0,
                punctuation_space=0.0,
                rgb_hex='000000',
                tag='default',
            )
            render_basic.compile_tex_file(tag='default')
            not_supported_chars.update(render_basic.get_not_supported_chars(tag='default'))

            begin = end

        return font_file, not_supported_chars, False

    except Exception:
        return font_file, set(chars), True


def proc_find_freetype_default_supported_chars(args: Tuple[str, List[str], str]):
    font_file, chars, _ = args

    face = freetype.Face(str(font_file), index=0)
    load_char_flags = freetype.FT_LOAD_RENDER  # type: ignore
    load_char_flags |= freetype.FT_LOAD_FORCE_AUTOHINT  # type: ignore

    base_res = 72
    font_size = 12
    face.set_char_size(
        width=font_size << 6,
        height=0,
        hres=base_res,
        vres=base_res,
    )

    not_supported_chars: Set[str] = set()
    for char in chars:
        not_supported = False
        try:
            face.load_char(char, load_char_flags)
            bitmap = face.glyph.bitmap
            width = bitmap.width
            if bitmap.rows == 0 or bitmap.width == 0 or width != bitmap.pitch:
                not_supported = True
        except Exception:
            logger.exception('Failed to load_char.')
        if not_supported:
            logger.warning(f'{font_file}: not supported char = {char}')
            not_supported_chars.add(char)

    return font_file, not_supported_chars, False


def proc_find_freetype_lcd_supported_chars(args: Tuple[str, List[str], str]):
    font_file, chars, _ = args

    face = freetype.Face(str(font_file), index=0)
    load_char_flags = freetype.FT_LOAD_RENDER  # type: ignore
    load_char_flags |= freetype.FT_LOAD_FORCE_AUTOHINT  # type: ignore
    load_char_flags |= freetype.FT_LOAD_TARGET_LCD  # type: ignore

    base_res = 72
    font_size = 12
    face.set_char_size(
        width=font_size << 6,
        height=0,
        hres=base_res,
        vres=base_res,
    )

    not_supported_chars: Set[str] = set()
    for char in chars:
        not_supported = False
        try:
            face.load_char(char, load_char_flags)
            bitmap = face.glyph.bitmap
            height = bitmap.rows
            pitch = bitmap.pitch
            if bitmap.rows == 0 or bitmap.width == 0 or len(bitmap.buffer) != height * pitch:
                not_supported = True
        except Exception:
            logger.exception('Failed to load_char.')
        if not_supported:
            logger.warning(f'{font_file}: not supported char = {char}')
            not_supported_chars.add(char)

    return font_file, not_supported_chars, False


def proc_find_fonttools_supported_chars(args: Tuple[str, List[str], str]):
    font_file, chars, _ = args

    if font_file.lower().endswith('ttc'):
        font = TTFont(font_file, fontNumber=0)
    else:
        font = TTFont(font_file)

    not_supported_chars: Set[str] = set()
    for char in chars:
        char_is_supported = False
        for cmap in font['cmap'].tables:  # type: ignore
            if cmap.isUnicode() and ord(char) in cmap.cmap:
                char_is_supported = True
                break
        if not char_is_supported:
            not_supported_chars.add(char)

    return font_file, not_supported_chars, False


def find_supported_chars_based_on_proc_func(
    proc_func: Callable[[Tuple[str, List[str], str]], Tuple[str, Set[str], bool]],
    font_folder: str,
    lexicons_json: str,
    render_folder: str,
    output: str,
):
    chars: List[str] = []
    for lexicon in LexiconCollection.from_file(lexicons_json).lexicons:
        chars.append(lexicon.char)
        chars.extend(lexicon.aliases)

    results = []
    with multiprocessing.Pool() as pool:
        for font_file, not_supported_chars, font_invalid in tqdm(
            pool.imap_unordered(
                proc_func, [(str(font_file), chars, render_folder)
                            for font_file, _ in search_font_files(font_folder)]
            )
        ):
            if font_invalid:
                logger.warning(f'Invalid font_file={font_file}, skip')
                continue
            results.append((font_file, not_supported_chars))

    out = []
    for font_file, not_supported_chars in results:
        supported_chars = []
        for char in chars:
            if char not in not_supported_chars:
                supported_chars.append(char)
        out.append({
            'font_file': str(font_file),
            'supported_chars': supported_chars,
        })

    io.write_json_lines(io.file(output), out, ensure_ascii=False)


def find_xelatex_supported_chars(
    font_folder: str,
    lexicons_json: str,
    render_folder: str,
    output: str,
):
    find_supported_chars_based_on_proc_func(
        proc_func=proc_find_xelatex_supported_chars,
        font_folder=font_folder,
        lexicons_json=lexicons_json,
        render_folder=render_folder,
        output=output,
    )


def find_freetype_default_supported_chars(
    font_folder: str,
    lexicons_json: str,
    render_folder: str,
    output: str,
):
    find_supported_chars_based_on_proc_func(
        proc_func=proc_find_freetype_default_supported_chars,
        font_folder=font_folder,
        lexicons_json=lexicons_json,
        render_folder=render_folder,
        output=output,
    )


def find_freetype_lcd_supported_chars(
    font_folder: str,
    lexicons_json: str,
    render_folder: str,
    output: str,
):
    find_supported_chars_based_on_proc_func(
        proc_func=proc_find_freetype_lcd_supported_chars,
        font_folder=font_folder,
        lexicons_json=lexicons_json,
        render_folder=render_folder,
        output=output,
    )


def find_fonttools_supported_chars(
    font_folder: str,
    lexicons_json: str,
    render_folder: str,
    output: str,
):
    find_supported_chars_based_on_proc_func(
        proc_func=proc_find_fonttools_supported_chars,
        font_folder=font_folder,
        lexicons_json=lexicons_json,
        render_folder=render_folder,
        output=output,
    )


def combine_supported_chars_jsls(supported_chars_jsls: str, output: str):
    supported_chars_jsl_list = supported_chars_jsls.split(',')
    font_file_to_supported_chars_sets = defaultdict(list)
    for supported_chars_jsl in supported_chars_jsl_list:
        for line in io.read_json_lines(supported_chars_jsl):
            font_file = line['font_file']
            supported_chars = line['supported_chars']
            font_file_to_supported_chars_sets[font_file].append(set(supported_chars))

    delta_log_pairs = []
    out = []
    for font_file, supported_chars_sets in font_file_to_supported_chars_sets.items():
        if len(supported_chars_sets) < len(supported_chars_jsl_list):
            print(f'!!! ignore {font_file}')
            continue
        union_supported_chars = set.union(*supported_chars_sets)
        intersection_supported_chars = set.intersection(*supported_chars_sets)
        delta = union_supported_chars - intersection_supported_chars
        log = f'{font_file}: delta = {len(delta)}, supported = {len(intersection_supported_chars)}'
        logger.info(log)
        delta_log_pairs.append((len(delta), log))
        out.append({
            'font_file': str(font_file),
            'supported_chars': sorted(intersection_supported_chars),
        })

    delta_log_pairs = sorted(delta_log_pairs, key=lambda p: p[0], reverse=True)
    logger.info('=' * 20)
    for _, log in delta_log_pairs[:20]:
        logger.info(log)

    io.write_json_lines(io.file(output), out, ensure_ascii=False)


def get_font_ratios(font_file: str, chars: Set[str]):
    font_size = 64
    char = '永'
    if char not in chars:
        return None, None, None

    font_face = freetype.Face(font_file)
    font_face.set_char_size(width=font_size * 64)

    load_char_flags = freetype.FT_LOAD_RENDER  # type: ignore
    font_face.load_char(char, load_char_flags)

    glyph = font_face.glyph
    bitmap = glyph.bitmap

    height = bitmap.rows
    width = bitmap.width
    assert width == bitmap.pitch

    # (H, W), [0, 255]
    np_image = np.array(bitmap.buffer, dtype=np.uint8).reshape(height, width)

    ascent = glyph.bitmap_top

    assert glyph.advance.y == 0
    pad_up = 0
    pad_down = 0

    # "bitmap’s left bearing"
    pad_left = glyph.bitmap_left
    # "It is always zero for horizontal layouts, and zero for vertical ones."
    assert glyph.advance.x > 0
    pad_right = round(glyph.advance.x / 64) - pad_left - width

    # NOTE: `bitmap_left` could be negative, we simply ignore the negative value.
    pad_left = max(0, pad_left)
    pad_right = max(0, pad_right)

    np_image, pad_up_inc, pad_down_inc = trim_char_np_image_vert(np_image)
    pad_up += pad_up_inc
    ascent -= pad_up_inc
    pad_down += pad_down_inc

    np_image, pad_left_inc, pad_right_inc = trim_char_np_image_hori(np_image)
    pad_left += pad_left_inc
    pad_right += pad_right_inc

    ascent_plus_pad_up_min_to_font_size_ratio = (ascent + pad_up) / font_size
    height_min_to_font_size_ratio = (pad_up + np_image.shape[0] + pad_down) / font_size
    width_min_to_font_size_ratio = (pad_left + np_image.shape[1] + pad_right) / font_size
    return (
        ascent_plus_pad_up_min_to_font_size_ratio,
        height_min_to_font_size_ratio,
        width_min_to_font_size_ratio,
    )


def build_font_collection(supported_chars_jsl: str, font_file_prefix: str, output_folder: str):
    font_metas: List[FontMeta] = []

    blacklist = [
        # FZYingBiXingShu-S16T transform simplified char to traditional char.
        'FZYingBiXingShu-S16T',
        'FZYingBiXingShu-S16S',
        # Japanese.
        'AquaKana',
        # Hard to read.
        'ZhiMangXing-Regular',
        'LiuJianMaoCao-Regular',
        # Too fancy.
        'MaShanZheng-Regular',
        'LongCang-Regular',
        'STXingkai',
        'STCaiyun',
        'STHupo',
        'fzmmt',
        # Low quality.
        '方正点阵冯简体',
        # cwTeX old version.
        'cwTeXQFangsong-Medium',
        'cwTeXQHei-Bold',
        'cwTeXQKai-Medium',
        'cwTeXQMing-Medium',
        'cwTeXQYuan-Medium',
        # Too fancy.
        'FZSTK',
        # 王汉宗字体系列.
        '注音',
        '破音',
        '鋼體',
        '波卡體',
        '空陰',
        '實陰',
        '半天水',
        '新潮體',
        '王漢宗標楷體空心',
        '勘亭',
    ]
    num_chars_min = 3000

    font_file_and_chars_pairs = []
    for line in io.read_json_lines(supported_chars_jsl):
        font_file = line['font_file']
        chars = line['supported_chars']

        hit_blacklist = None
        for name in blacklist:
            if name in font_file:
                hit_blacklist = name
                break
        if hit_blacklist:
            logger.info(f'{font_file} blacklist {hit_blacklist}, skip.')
            continue

        if len(chars) < num_chars_min:
            logger.info(f'{font_file} {len(chars)} chars < {num_chars_min}, skip.')
            continue

        font_file_and_chars_pairs.append((font_file, chars))

    font_file_and_chars_pairs = sorted(font_file_and_chars_pairs, key=lambda p: len(p[1]))
    ttf_pairs = []
    otf_pairs = []
    ttc_pairs = []
    for font_file, chars in font_file_and_chars_pairs:
        pair = (font_file, chars)
        if font_file.lower().endswith('ttf'):
            ttf_pairs.append(pair)
        elif font_file.lower().endswith('otf'):
            otf_pairs.append(pair)
        elif font_file.lower().endswith('ttc'):
            ttc_pairs.append(pair)
        else:
            raise NotImplementedError()

    ttf_prefixes = [
        'Deng',
        'GenJyuuGothicL',
        'GenRyuMinTW',
        'GenShinGothic',
        'GenWanMinTW',
        'GenYoMinTW',
        'OPPOSans',
        'Microsoft-Yahei',
        'SourceHanSerifCN',
    ]
    ttf_prefix_to_ttf_pairs = defaultdict(list)
    other_ttf_pairs = []
    for font_file, chars in ttf_pairs:
        name = font_file.split('/')[-1]
        hit_ttf_prefix = None
        for ttf_prefix in ttf_prefixes:
            if name.startswith(ttf_prefix):
                hit_ttf_prefix = ttf_prefix
                break
        if hit_ttf_prefix:
            ttf_prefix_to_ttf_pairs[hit_ttf_prefix].append((font_file, chars))
        else:
            other_ttf_pairs.append((font_file, chars))
    assert len(ttf_prefixes) == len(ttf_prefix_to_ttf_pairs)

    for ttf_prefix, ttf_pairs in ttf_prefix_to_ttf_pairs.items():
        assert len(ttf_pairs) > 1
        logger.info(
            f'ttf_prefix={ttf_prefix}, {[io.file(font_file).stem for font_file, _ in ttf_pairs]}'
        )

        # Assert the same charset.
        supported_chars_sets = [set(chars) for _, chars in ttf_pairs]
        union_supported_chars = set.union(*supported_chars_sets)
        intersection_supported_chars = set.intersection(*supported_chars_sets)
        if union_supported_chars != intersection_supported_chars:
            logger.warning(f'!!! VTTC chars conflict, ttf_prefix={ttf_prefix}')

        (
            ascent_plus_pad_up_min_to_font_size_ratio,
            height_min_to_font_size_ratio,
            width_min_to_font_size_ratio,
        ) = get_font_ratios(ttf_pairs[0][0], union_supported_chars)
        if ascent_plus_pad_up_min_to_font_size_ratio is None:
            continue
        assert height_min_to_font_size_ratio is not None
        assert width_min_to_font_size_ratio is not None

        font_metas.append(
            FontMeta(
                name=ttf_prefix,
                kind=FontKind.VTTC,
                chars=sorted(intersection_supported_chars),
                font_files=[font_file for font_file, _ in ttf_pairs],
                ascent_plus_pad_up_min_to_font_size_ratio=ascent_plus_pad_up_min_to_font_size_ratio,
                height_min_to_font_size_ratio=height_min_to_font_size_ratio,
                width_min_to_font_size_ratio=width_min_to_font_size_ratio,
            )
        )

    for font_file, chars in other_ttf_pairs:
        (
            ascent_plus_pad_up_min_to_font_size_ratio,
            height_min_to_font_size_ratio,
            width_min_to_font_size_ratio,
        ) = get_font_ratios(font_file, set(chars))
        if ascent_plus_pad_up_min_to_font_size_ratio is None:
            continue
        assert height_min_to_font_size_ratio is not None
        assert width_min_to_font_size_ratio is not None

        name = io.file(font_file).stem
        font_metas.append(
            FontMeta(
                name=name,
                kind=FontKind.VTTC,
                chars=chars,
                font_files=[font_file],
                ascent_plus_pad_up_min_to_font_size_ratio=ascent_plus_pad_up_min_to_font_size_ratio,
                height_min_to_font_size_ratio=height_min_to_font_size_ratio,
                width_min_to_font_size_ratio=width_min_to_font_size_ratio,
            )
        )

    del ttf_pairs
    del ttf_prefix_to_ttf_pairs
    del other_ttf_pairs

    otf_prefixes = [
        'AlibabaPuHuiTi',
        'DottedSongti',
        'NotoSansSC',
        'NotoSansTC',
        'NotoSerifSC',
        'NotoSerifTC',
        'SourceHanSansCN',
    ]
    otf_prefix_to_otf_pairs = defaultdict(list)
    other_otf_pairs = []
    for font_file, chars in otf_pairs:
        name = font_file.split('/')[-1]
        hit_otf_prefix = None
        for otf_prefix in otf_prefixes:
            if name.startswith(otf_prefix):
                hit_otf_prefix = otf_prefix
                break
        if hit_otf_prefix:
            otf_prefix_to_otf_pairs[hit_otf_prefix].append((font_file, chars))
        else:
            other_otf_pairs.append((font_file, chars))
    assert len(otf_prefixes) == len(otf_prefix_to_otf_pairs)

    for otf_prefix, otf_pairs in otf_prefix_to_otf_pairs.items():
        assert len(otf_pairs) > 1
        logger.info(
            f'otf_prefix={otf_prefix}, {[io.file(font_file).stem for font_file, _ in otf_pairs]}'
        )

        # Assert the same charset.
        supported_chars_sets = [set(chars) for _, chars in otf_pairs]
        union_supported_chars = set.union(*supported_chars_sets)
        intersection_supported_chars = set.intersection(*supported_chars_sets)
        if union_supported_chars != intersection_supported_chars:
            logger.warning(f'!!! VTTC chars conflict, ttf_prefix={otf_prefix}')

        (
            ascent_plus_pad_up_min_to_font_size_ratio,
            height_min_to_font_size_ratio,
            width_min_to_font_size_ratio,
        ) = get_font_ratios(otf_pairs[0][0], set(union_supported_chars))
        if ascent_plus_pad_up_min_to_font_size_ratio is None:
            continue
        assert height_min_to_font_size_ratio is not None
        assert width_min_to_font_size_ratio is not None

        font_metas.append(
            FontMeta(
                name=otf_prefix,
                kind=FontKind.VOTC,
                chars=sorted(intersection_supported_chars),
                font_files=[font_file for font_file, _ in otf_pairs],
                ascent_plus_pad_up_min_to_font_size_ratio=ascent_plus_pad_up_min_to_font_size_ratio,
                height_min_to_font_size_ratio=height_min_to_font_size_ratio,
                width_min_to_font_size_ratio=width_min_to_font_size_ratio,
            )
        )

    for font_file, chars in other_otf_pairs:
        (
            ascent_plus_pad_up_min_to_font_size_ratio,
            height_min_to_font_size_ratio,
            width_min_to_font_size_ratio,
        ) = get_font_ratios(font_file, set(chars))
        if ascent_plus_pad_up_min_to_font_size_ratio is None:
            continue
        assert height_min_to_font_size_ratio is not None
        assert width_min_to_font_size_ratio is not None

        name = io.file(font_file).stem
        font_metas.append(
            FontMeta(
                name=name,
                kind=FontKind.VOTC,
                chars=chars,
                font_files=[font_file],
                ascent_plus_pad_up_min_to_font_size_ratio=ascent_plus_pad_up_min_to_font_size_ratio,
                height_min_to_font_size_ratio=height_min_to_font_size_ratio,
                width_min_to_font_size_ratio=width_min_to_font_size_ratio,
            )
        )

    del otf_pairs
    del otf_prefix_to_otf_pairs
    del other_otf_pairs

    for font_file, chars in ttc_pairs:
        (
            ascent_plus_pad_up_min_to_font_size_ratio,
            height_min_to_font_size_ratio,
            width_min_to_font_size_ratio,
        ) = get_font_ratios(font_file, set(chars))
        if ascent_plus_pad_up_min_to_font_size_ratio is None:
            continue
        assert height_min_to_font_size_ratio is not None
        assert width_min_to_font_size_ratio is not None

        name = io.file(font_file).stem
        font_metas.append(
            FontMeta(
                name=name,
                kind=FontKind.TTC,
                chars=chars,
                font_files=[font_file],
                ttc_font_index_max=get_font_index_max(str(font_file)),
                ascent_plus_pad_up_min_to_font_size_ratio=ascent_plus_pad_up_min_to_font_size_ratio,
                height_min_to_font_size_ratio=height_min_to_font_size_ratio,
                width_min_to_font_size_ratio=width_min_to_font_size_ratio,
            )
        )

    del ttc_pairs

    added_names = set()
    added_font_files = set()
    for font_meta in font_metas:
        assert font_meta.name not in added_names
        added_names.add(font_meta.name)
        for font_file in font_meta.font_files:
            assert font_file not in added_font_files
            added_font_files.add(font_file)

    out_fd = io.folder(output_folder, reset=True)
    font_fd = io.folder(out_fd / FontCollectionFolderTree.FONT.value, touch=True)
    font_meta_fd = io.folder(out_fd / FontCollectionFolderTree.FONT_META.value, touch=True)

    for font_meta in font_metas:
        logger.info(f'Saving {font_meta}')
        font_meta.to_file(
            font_meta_fd / f'{font_meta.name}.json',
            font_file_prefix=font_file_prefix,
        )
        for font_file in font_meta.font_files:
            font_file = io.file(font_file, exists=True)
            shutil.copyfile(font_file, font_fd / font_file.name)


def build_open_source_font_collection(
    font_collection_folder: str,
    open_source_font_flatten_folder: str,
    output_folder: str,
):
    font_collection = FontCollection.from_folder(font_collection_folder)
    open_source_file_names = set(
        file_name for _, file_name in search_font_files(open_source_font_flatten_folder)
    )
    open_source_font_metas = []
    for font_meta in font_collection.font_metas:
        num_hits = 0
        for font_file in font_meta.font_files:
            font_file = io.file(font_file)
            if font_file.name in open_source_file_names:
                num_hits += 1
        if num_hits > 0:
            assert num_hits == len(font_meta.font_files)
            open_source_font_metas.append(font_meta)

    out_fd = io.folder(output_folder, reset=True)
    font_fd = io.folder(out_fd / FontCollectionFolderTree.FONT.value, touch=True)
    font_meta_fd = io.folder(out_fd / FontCollectionFolderTree.FONT_META.value, touch=True)

    for font_meta in open_source_font_metas:
        logger.info(f'Saving {font_meta}')
        font_meta.to_file(
            font_meta_fd / f'{font_meta.name}.json',
            font_file_prefix=io.folder(font_collection_folder)
            / FontCollectionFolderTree.FONT.value,
        )
        for font_file in font_meta.font_files:
            font_file = io.file(font_file, exists=True)
            shutil.copyfile(font_file, font_fd / font_file.name)
