from typing import Dict, Set, Tuple, List, Optional
import re
from itertools import chain
import string

import iolite as io
import cattrs
import opencc

from vkit.utility.text import LexiconType, get_lexicon_type, normalize
from vkit.element import Lexicon, LexiconCollection


def _match_tygfhzb_lexicons(txt_file: str):
    text = io.file(txt_file, exists=True).read_text()
    for match in re.finditer(r'^:(\d+)\s(.+?)$', text, re.UNICODE | re.MULTILINE):
        idx = int(match.group(1))
        char: str = match.group(2)

        if len(char) > 1:
            src = char
            if char.startswith('{{!|') and char.endswith('}}'):
                char = char[4:-2].strip()
                char = char[0]
                print(f'{idx}: {src} -> {char}')
            elif char[0] != '{':
                char = char[0]
                print(f'{idx}: {src} -> {char}')
            else:
                raise NotImplementedError()

        char = normalize(char)
        assert len(char) == 1
        yield idx, char


def extract_lexicons_from_tygfhzb(
    level1_txt: str,
    level2_txt: str,
    level3_txt: str,
    output_json: str,
):
    idx_char_pairs: List[Tuple[int, str]] = []
    idx_char_pairs.extend(_match_tygfhzb_lexicons(level1_txt))
    idx_char_pairs.extend(_match_tygfhzb_lexicons(level2_txt))
    idx_char_pairs.extend(_match_tygfhzb_lexicons(level3_txt))

    expecte_num_chars = 8105

    # No missing idx.
    idx_to_char = dict(idx_char_pairs)
    for idx in range(1, expecte_num_chars + 1):
        assert idx in idx_to_char

    # No duplicated chars.
    assert len(set(idx_to_char.values())) == expecte_num_chars

    io.write_json(
        output_json,
        idx_to_char,
        indent=2,
        ensure_ascii=False,
    )


def _normalize_group(line: str):
    pattern = r'\{\{!\|(.+?)\}\}'
    matches = list(re.finditer(pattern, line, re.UNICODE | re.MULTILINE))
    matches = sorted(matches, key=lambda m: m.start())
    segments: List[str] = []
    prev_end = 0
    for match in matches:
        begin = match.start()
        if prev_end < begin:
            segments.append(line[prev_end:begin])
        end = match.end()
        group = match.group(1)
        char = group[0]
        print(f'GROUP: {group} -> {char}')
        segments.append(char)
        prev_end = end
    if prev_end < len(line):
        segments.append(line[prev_end:])
    return ''.join(segments)


def _normalize_tag(line: str):
    pattern = r'\|.+?=\".+?\"\|'
    new_line = re.sub(pattern, '||', line)
    return new_line


def _line_to_row(line: str):
    line = line.strip()
    src = line
    assert src

    if line[0] == '|' and line[1] != '|':
        line = line[1:]
    elif line.startswith('||') and line[2] != '|':
        line = line[2:]
    elif line.startswith('|||'):
        line = line[1:]

    new_line = re.sub(r'\{\{.+?\}\}', '', line)
    if new_line != line:
        print(f'DROP GORUP: {line} -> {new_line}')
        line = new_line

    new_line = re.sub(r'\[\[.+?\]\]', '', line)
    if new_line != line:
        print(f'DROP GORUP: {line} -> {new_line}')
        line = new_line

    row: List[Optional[str]] = []
    begin = 0
    end = 0
    while end < len(line):
        if line[end] != '|':
            end += 1
            continue

        assert end + 1 < len(line) and line[end + 1] == '|'
        if begin == end:
            row.append(None)
        else:
            segment = line[begin:end]
            row.append(segment)

        begin = end + 2
        end = end + 2

    if begin < len(line):
        row.append(line[begin:])

    for segment in row:
        if segment:
            assert '|' not in segment

    return row


def _process_trads_or_variants(texts: List[str]):
    processed: List[str] = []
    for text in texts:
        text = normalize(text)
        text = ''.join(char for char in text if char not in ('(', ')', '[', ']', '~'))

        if len(text) > 1:
            pattern = r'&.+?;'
            new_text = re.sub(pattern, '', text)
            if new_text != text:
                print(f'PATCH: {text} -> {new_text}')
                text = new_text

            pattern = r'<!--.+?-->'
            new_text = re.sub(pattern, '', text)
            if new_text != text:
                print(f'PATCH: {text} -> {new_text}')
                text = new_text

            print(f'WARN: _process_trads_or_variants: {text}')

        if not text:
            continue

        text = ''.join(set(text))
        for char in text:
            assert get_lexicon_type(char) == LexiconType.CHINESE

        processed.extend(text)
    return processed


def extract_lexicons_from_tygfhzb_appendix(
    appendix_txt: str,
    output_json: str,
):
    text = io.file(appendix_txt, exists=True).read_text()

    rows: List[List[Optional[str]]] = []

    pattern = r'\{\{col\-begin\}\}(.+?)\{\{col\-end\}\}'
    for match_table in re.finditer(pattern, text, re.UNICODE | re.MULTILINE | re.DOTALL):
        body: str = match_table.group(1)
        lines = body.splitlines()
        for line in lines:
            src = line
            assert src
            line = line.strip()
            if not line.startswith('|'):
                continue
            if line in ('|-', '|}'):
                continue

            line = _normalize_group(line)
            line = _normalize_tag(line)
            row = _line_to_row(line)
            print('>', line)
            print('---', row)
            rows.append(row)

    # Grouping.
    groups: List[Tuple[int, str, bool, List[str], List[str]]] = []
    begin = 0
    end = 0
    while end <= len(rows):
        is_num = False
        if end < len(rows):
            row = rows[end]
            if row[0]:
                try:
                    int(row[0])  # type: ignore
                    is_num = True
                except ValueError:
                    pass
        else:
            is_num = True

        if 0 < end and is_num:
            idx = int(rows[begin][0])  # type: ignore

            assert rows[begin][1]
            char = normalize(rows[begin][1])  # type: ignore
            assert len(char) == 1

            trads: List[str] = []
            variants: List[str] = []
            if len(rows[begin]) >= 3 and rows[begin][2]:
                trads.append(rows[begin][2])  # type: ignore
            if len(rows[begin]) >= 4 and rows[begin][3]:
                variants.append(rows[begin][3])  # type: ignore

            for row_idx in range(begin + 1, end):
                row = rows[row_idx]
                if len(row) <= 2:
                    if row[0]:
                        trads.append(row[0])
                    if len(row) == 2 and row[1]:
                        variants.append(row[1])
                else:
                    assert len(row) <= 4
                    assert not row[0] and not row[1]
                    if row[2]:
                        trads.append(row[2])
                    if len(row) == 4 and row[3]:
                        variants.append(row[3])

            if groups:
                assert groups[-1][0] < idx
            assert 1 <= idx <= 8105

            # if idx == 438:
            #     breakpoint()

            has_tilde = ('~' in trads) or ('～' in trads)
            trads = _process_trads_or_variants(trads)
            variants = _process_trads_or_variants(variants)
            groups.append((idx, char, has_tilde, trads, variants))
            begin = end

        end += 1

    assert len(groups) == 3120
    io.write_json(
        output_json,
        [{
            'index': idx,
            'char': char,
            'tilded': tilded,
            'trads': trads,
            'variants': variants,
        } for idx, char, tilded, trads, variants in groups],
        indent=2,
        ensure_ascii=False,
    )


def combine_tygfhzb(
    raw_tygfhzb_json: str,
    raw_tygfhzb_appendix_json: str,
    blacklist_chars_json: str,
    output_json: str,
):
    idx_to_char: Dict[int, str] = cattrs.structure(io.read_json(raw_tygfhzb_json), Dict[int, str])

    blacklist_chars: Set[str] = set(io.read_json(blacklist_chars_json))
    groups = io.read_json(raw_tygfhzb_appendix_json)
    for group in groups:
        assert group['char'] not in blacklist_chars
        trads = [char for char in group['trads'] if char not in blacklist_chars]
        group['trads'] = trads
        variants = [char for char in group['variants'] if char not in blacklist_chars]
        group['variants'] = variants

    idx_to_group: Dict[int, Dict] = {group['index']: group for group in groups}

    # assert idx_to_group[332]['char'] == '奶'
    # idx_to_group[332]['variants'].pop(0) == '妳'

    # assert idx_to_group[1829]['char'] == '恶'
    # idx_to_group[1829]['trads'].pop(1) == '噁'

    # assert idx_to_group[1639]['char'] == '阁'
    # idx_to_group[1639]['variants'].pop(0) == '閤'

    # assert idx_to_group[675]['char'] == '克'
    # idx_to_group[675]['variants'].pop(0) == '尅'

    visited_trads: Set[str] = set()
    visited_variants: Set[str] = set()
    for group in groups:
        idx = group['index']
        char = group['char']
        assert idx_to_char[idx] == char

        for char in group['trads']:
            if char in visited_trads:
                print('CONFLICT TRADS', char)
            visited_trads.add(char)
        for char in group['variants']:
            if char in visited_variants:
                print('CONFLICT VARIANTS', char)
            visited_variants.add(char)

    combined_groups = []
    for idx, char in sorted(idx_to_char.items(), key=lambda p: p[0]):
        if idx in idx_to_group:
            group = idx_to_group[idx]
            assert group['char'] == char
            combined_groups.append(group)
        else:
            combined_groups.append({
                'index': idx,
                'char': char,
                'tilded': False,
                'trads': [],
                'variants': [],
            })

    all_chars = set()
    std_and_trad_chars = set()
    for group in combined_groups:
        all_chars.add(group['char'])
        std_and_trad_chars.add(group['char'])
        all_chars.update(group['trads'])
        std_and_trad_chars.update(group['trads'])
        all_chars.update(group['variants'])
    print('STD', len(combined_groups))
    print(
        'STD + TRADS', len(std_and_trad_chars), 'DELTA',
        len(std_and_trad_chars) - len(combined_groups)
    )
    print(
        'STD + TRADS + VARIANTS', len(all_chars), 'DELTA',
        len(all_chars) - len(std_and_trad_chars)
    )

    io.write_json(
        output_json,
        combined_groups,
        indent=2,
        ensure_ascii=False,
    )


def more_chinese_chars(tygfhzb_json: str, too_close_txt: str, output_json: str):
    groups = io.read_json(tygfhzb_json)

    current_all_chars: Set[str] = set()
    for group in groups:
        current_all_chars.add(group['char'])
        current_all_chars.update(group['trads'])
        current_all_chars.update(group['variants'])

    new_chars: Set[str] = set()
    blacklist_chars: Set[str] = set()

    for line in io.read_text_lines(too_close_txt, strip=True, skip_empty=True):
        chars = [normalize(char) for char in line.split()]
        if chars[0] not in current_all_chars:
            print('Add', chars[0])
            new_chars.add(chars[0])
        blacklist_chars.update(chars[1:])

    # Manually added.
    assert '〇' not in current_all_chars
    new_chars.add('〇')

    opencc_s2t_cvt = opencc.OpenCC('s2t.json')
    for group in groups:
        char = group['char']
        new_char = opencc_s2t_cvt.convert(char)
        if len(char) != 1:
            print(f'WARN: {char} -> {new_char}')
            continue
        new_char = normalize(new_char)
        if new_char not in current_all_chars and new_char not in blacklist_chars:
            new_chars.add(new_char)

    if '𣈶' in new_chars:
        new_chars.remove('𣈶')

    print('NEW CHARS:', len(new_chars))
    io.write_json(
        output_json,
        sorted(new_chars),
        indent=2,
        ensure_ascii=False,
    )


def generate_chinese_lexicon_collection(
    tygfhzb_json: str,
    more_chinese_chars_json: str,
    output_json: str,
):
    groups = io.read_json(tygfhzb_json)
    more_chinese_chars = io.read_json(more_chinese_chars_json)

    level_1_stds = {group['char'] for group in groups if group['index'] <= 3500}
    level_2_stds = {group['char'] for group in groups if 3500 < group['index'] <= 6500}
    level_3_stds = {group['char'] for group in groups if 6500 < group['index']}
    tilded_stds = {group['char'] for group in groups if group['tilded']}
    trads = set(chain.from_iterable(group['trads'] for group in groups))
    variants = set(chain.from_iterable(group['variants'] for group in groups))

    chars: Set[str] = set()
    for group in groups:
        for char in chain([group['char']], group['trads'], group['variants']):
            assert normalize(char) == char and len(char) == 1
            chars.add(char)

    added_chars: Set[str] = set()
    lexicons: List[Lexicon] = []
    for char in chars:
        tags: List[str] = []
        if char in level_1_stds:
            tags.append('chinese_std_level_1')
        if char in level_2_stds:
            tags.append('chinese_std_level_2')
        if char in level_3_stds:
            tags.append('chinese_std_level_3')
        if char in tilded_stds:
            tags.append('chinese_std_tilded')
        if char in trads:
            tags.append('chinese_traditional')
        if char in variants:
            tags.append('chinese_variant')

        assert char not in added_chars
        added_chars.add(char)
        lexicons.append(Lexicon(
            char=char,
            aliases=[],
            tags=tags,
        ))

    for char in more_chinese_chars:
        assert normalize(char) == char and len(char) == 1
        assert char not in added_chars
        added_chars.add(char)

        if char == '〇':
            tags = ['chinese_std_level_1']
        else:
            tags = ['chinese_traditional_extended']

        lexicons.append(Lexicon(
            char=char,
            aliases=[],
            tags=tags,
        ))

    lexicons = sorted(lexicons, key=lambda lexicon: tuple(lexicon.tags))
    LexiconCollection(lexicons=lexicons).to_file(output_json)


def _to_full_width(char: str):
    assert 0x21 <= ord(char) <= 0x7E
    return chr(ord(char) + 0xFEE0)


def generate_english_delimiter_lexicon_collection(output_json: str):
    lexicons: List[Lexicon] = []
    for char in string.punctuation:
        assert normalize(char) == char and len(char) == 1

        char_full_width = _to_full_width(char)
        assert char_full_width != char

        lexicons.append(Lexicon(
            char=char,
            aliases=[char_full_width],
            tags=['english_delimiter'],
        ))

    lexicons = sorted(lexicons, key=lambda lexicon: tuple(lexicon.tags))
    LexiconCollection(lexicons=lexicons).to_file(output_json)


def generate_chinese_english_delimiter_lexicon_collection(
    english_delimiter_lexicon_collection_json: str,
    output_json: str,
):
    english_lexicon_collection = LexiconCollection.from_file(
        english_delimiter_lexicon_collection_json
    )
    english_lexicons = english_lexicon_collection.lexicons
    char_to_english_lexicon = {lexicon.char: lexicon for lexicon in english_lexicons}
    all_english_chars = set(
        chain.from_iterable(lexicon.char_and_aliases for lexicon in english_lexicons)
    )

    # FORM https://zh.wikipedia.org/wiki/%E6%A0%87%E7%82%B9%E7%AC%A6%E5%8F%B7
    chinese_english_lexicons: List[Lexicon] = []
    for new_chars in [
        ['¢', '￠'],
        ['£', '￡'],
        ['¥', '￥'],
        ['‰'],
        ['‱'],
        ['※'],
        ['§'],
        ['†'],
        ['…', '⋯', '᠁'],
        ['、'],
        ['。'],
        ['《'],
        ['》'],
        # NOTE: 在此我们不区分内部与中空引号
        ['「', '『'],
        ['」', '』'],
        # NOTE: 竖排引号
        ['﹁', '﹃'],
        ['﹂', '﹄'],
    ]:
        assert len(set(new_chars)) == len(new_chars)
        assert all(len(char) == 1 for char in new_chars)
        assert all(char not in all_english_chars for char in new_chars)
        char = new_chars[0]
        aliases = new_chars[1:]
        chinese_english_lexicons.append(
            Lexicon(
                char=char,
                aliases=aliases,
                tags=['chinese_delimiter'],
            )
        )

    english_char_to_chinese_aliases = {
        # 间隔号：
        # https://zh.wikipedia.org/wiki/%E9%97%B4%E9%9A%94%E5%8F%B7
        # NOTE: 项目编号 “•” 从形态上难以与前两者区分，所以也编入此组.
        '.': ['·', '‧', '•'],
        '-': ['—'],
        "'": ['‘', '’', '′'],
        '"': ['“', '”', '〃'],
        '<': ['〈'],
        '>': ['〉'],
        # NOTE: 中空方括号、六角括号不常用，因此归为实心方括号的别名.
        '[': ['【', '〖', '〔'],
        # NOTE: 中空方括号不常用，因此归为实心方括号的别名
        ']': ['】', '〗', '〕'],
    }
    for english_char, chinese_aliases in english_char_to_chinese_aliases.items():
        assert english_char not in chinese_aliases
        assert len(set(chinese_aliases)) == len(chinese_aliases)
        english_lexicon = char_to_english_lexicon[english_char]
        combined_aliases = list(english_lexicon.aliases) + chinese_aliases
        assert len(set(combined_aliases)) == len(english_lexicon.aliases) + len(chinese_aliases)
        extended_tags = list(english_lexicon.tags) + ['chinese_delimiter_extended']
        chinese_english_lexicons.append(
            Lexicon(
                char=english_char,
                aliases=combined_aliases,
                tags=extended_tags,
            )
        )

    added: Set[str] = set()
    for lexicon in chinese_english_lexicons:
        assert len(set(lexicon.char_and_aliases)) == len(lexicon.char_and_aliases)
        assert all(char not in added for char in lexicon.char_and_aliases)
        added.update(lexicon.char_and_aliases)

    for char in list(set(char_to_english_lexicon) - added):
        lexicon = char_to_english_lexicon[char]
        assert len(set(lexicon.char_and_aliases)) == len(lexicon.char_and_aliases)
        assert all(char not in added for char in lexicon.char_and_aliases)
        chinese_english_lexicons.append(lexicon)
        added.update(lexicon.char_and_aliases)

    chinese_english_lexicons = sorted(
        chinese_english_lexicons, key=lambda lexicon: tuple(lexicon.tags)
    )
    LexiconCollection(lexicons=chinese_english_lexicons).to_file(output_json)


def generate_english_lexicon_collection(output_json: str):
    lexicons: List[Lexicon] = []
    for char in string.ascii_letters:
        assert normalize(char) == char and len(char) == 1

        # Full width form as alias.
        char_full_width = _to_full_width(char)
        assert char_full_width != char

        lexicons.append(Lexicon(
            char=char,
            aliases=[char_full_width],
            tags=['english_std'],
        ))

        # Also add the circled form.
        if char.islower():
            circled_char = chr(ord(char) - ord('a') + 9424)
        else:
            circled_char = chr(ord(char) - ord('A') + 9398)
        print('CIRCLED FORM:', char, circled_char)

        lexicons.append(Lexicon(
            char=circled_char,
            aliases=[],
            tags=['english_circled'],
        ))

    lexicons = sorted(lexicons, key=lambda lexicon: tuple(lexicon.tags))
    LexiconCollection(lexicons=lexicons).to_file(output_json)


def generate_digit_lexicon_collection(output_json: str):
    lexicons: List[Lexicon] = []
    for char in string.digits:
        assert normalize(char) == char and len(char) == 1

        # Full width form as alias.
        char_full_width = _to_full_width(char)
        assert char_full_width != char

        lexicons.append(Lexicon(
            char=char,
            aliases=[char_full_width],
            tags=['digit_std'],
        ))

        # Also add the circled form.
        if char == '0':
            circled_chars = [chr(9450)]
        else:
            offset = ord(char) - ord('1')
            circled_chars = [chr(9312 + offset), chr(9461 + offset)]

        print('CIRCLED FORM:', char, circled_chars)
        lexicons.append(
            Lexicon(
                char=circled_chars[0],
                aliases=circled_chars[1:],
                tags=['digit_circled'],
            )
        )

    lexicons = sorted(lexicons, key=lambda lexicon: tuple(lexicon.tags))
    LexiconCollection(lexicons=lexicons).to_file(output_json)


def combine_lexicon_collections(json_files: str, output_json: str):
    combined_lexicons: List[Lexicon] = []
    added: Set[str] = set()

    for json_file in json_files.split(','):
        json_file = io.file(json_file, exists=True)
        lexicons = LexiconCollection.from_file(json_file).lexicons

        for lexicon in lexicons:
            assert len(set(lexicon.char_and_aliases)) == len(lexicon.char_and_aliases)
            assert all(char not in added for char in lexicon.char_and_aliases)
            combined_lexicons.append(lexicon)
            added.update(lexicon.char_and_aliases)

        print(f'{json_file.name} {len(lexicons)} merged.')

    print('ALL', len(combined_lexicons))
    LexiconCollection(lexicons=combined_lexicons).to_file(output_json)
